"""Utilities for validating results."""

__author__ = ["MatthewMiddlehurst"]
__all__ = [
    "validate_results_file",
    "compare_result_file_resample",
]


def validate_results_file(file_path):
    """Validate that a results file is in the correct format.

    Validates that the first, second, third and results lines follow the expected
    format. This does not verify that the actual contents of the results file make
    sense.

    Works for classification, regression and clustering results files.

    Parameters
    ----------
    file_path : str
        Path to the results file to be validated, including the file itself.

    Returns
    -------
    valid_file : bool
        True if the results file is valid, False otherwise.
    """
    with open(file_path) as f:
        lines = f.readlines()

    if not _check_first_line(lines[0]) or not _check_second_line(lines[1]):
        return False

    if _check_classification_third_line(lines[2]):
        n_probas = int(lines[2].split(",")[5])
        probabilities = True
    elif _check_clustering_third_line(lines[2]):
        n_probas = int(lines[2].split(",")[6])
        probabilities = True
    elif _check_regression_third_line(lines[2]) or _check_forecasting_third_line(
        lines[2]
    ):
        n_probas = 0
        probabilities = False
    else:
        return False

    for i in range(3, len(lines)):
        if not _check_results_line(
            lines[i], probabilities=probabilities, n_probas=n_probas
        ):
            return False

    return True


def _check_first_line(line):
    line = line.split(",")
    return len(line) >= 5


def _check_second_line(line):
    line = line.split(",")
    return len(line) >= 1


def _check_classification_third_line(line):
    line = line.split(",")
    floats = [0, 1, 2, 3, 4, 5, 7, 8]
    return _check_line_length_and_floats(line, 9, floats)


def _check_regression_third_line(line):
    line = line.split(",")
    floats = [0, 1, 2, 3, 4, 6, 7]
    return _check_line_length_and_floats(line, 8, floats)


def _check_clustering_third_line(line):
    line = line.split(",")
    floats = [0, 1, 2, 3, 4, 5, 6]
    return _check_line_length_and_floats(line, 7, floats)


def _check_forecasting_third_line(line):
    line = line.split(",")
    floats = [0, 1, 2, 3, 4]
    return _check_line_length_and_floats(line, 5, floats)


def _check_line_length_and_floats(line, length, floats):
    if len(line) != length:
        return False

    for i in floats:
        try:
            float(line[i])
        except ValueError:
            return False

    return True


def _check_results_lines(lines, num_results_lines=None, probabilities=True, n_probas=2):
    if num_results_lines is not None:
        assert len(lines) - 3 == num_results_lines

        for i in range(3, num_results_lines):
            assert _check_results_line(
                lines[i], probabilities=probabilities, n_probas=n_probas
            )
    else:
        for i in range(3, 6):
            assert _check_results_line(
                lines[i], probabilities=probabilities, n_probas=n_probas
            )


def _check_results_line(line, probabilities=True, n_probas=2):
    line = line.split(",")

    if len(line) < 2:
        return False

    try:
        float(line[0])
        float(line[1])
    except ValueError:
        return False

    if probabilities:
        if len(line) < 3 + n_probas or line[2] != "":
            return False

        psum = 0
        try:
            for i in range(n_probas):
                psum += float(line[3 + i])
        except ValueError:
            return False

        if psum < 0.99999 or psum > 1.00001:
            return False
    else:
        n_probas = 0

    if len(line) > 4 + n_probas:
        if line[4 + n_probas] != "":
            return False

        try:
            float(line[4 + n_probas])
        except ValueError:
            return False

    if len(line) > 5 + n_probas and line[5 + n_probas] != "":
        return False

    return True


def compare_result_file_resample(file_path1, file_path2):
    """Validate that two results files use the same data resample.

    Files are deemed as having the same resample if the file length is the same and all
    true label values are the same in both files.

    Parameters
    ----------
    file_path1 : str
        Path to the first results file to be compared, including the file itself.
    file_path1 : str
        Path to the second results file to be compared, including the file itself.

    Returns
    -------
    same_resample : bool
        True if the results file use the same data resample, False otherwise.
    """
    with open(file_path1) as f:
        lines1 = f.readlines()

    with open(file_path2) as f:
        lines2 = f.readlines()

    if len(lines1) != len(lines2):
        raise ValueError("Input results file have different numbers of lines.")

    for i in range(3, len(lines1)):
        if lines1[i].split(",")[0] != lines2[i].split(",")[0]:
            return False

    return True



"""Tests for publication experiments estimator selection."""

import pytest

from tsml_eval.publications.y2023.rist_pipeline import (
    _set_rist_classifier,
    _set_rist_regressor,
    rist_classifiers,
    rist_regressors,
)
from tsml_eval.testing.testing_utils import _check_set_method, _check_set_method_results


def test_set_rist_classifier():
    """Test set_rist_classifier method."""
    classifier_dict = {}
    all_classifier_names = []

    _check_set_method(
        _set_rist_classifier,
        rist_classifiers,
        classifier_dict,
        all_classifier_names,
    )

    _check_set_method_results(
        classifier_dict,
        estimator_name="Classifiers",
        method_name="_set_rist_classifier",
    )


def test_set_rist_classifier_invalid():
    """Test set_rist_classifierr method with invalid estimator."""
    with pytest.raises(ValueError, match="UNKNOWN CLASSIFIER"):
        _set_rist_classifier("invalid")


def test_set_rist_regressor():
    """Test set_rist_regressors method."""
    regressor_dict = {}
    all_regressor_names = []

    _check_set_method(
        _set_rist_regressor,
        rist_regressors,
        regressor_dict,
        all_regressor_names,
    )

    _check_set_method_results(
        regressor_dict, estimator_name="Regressors", method_name="_set_rist_regressor"
    )


def test_set_rist_regressor_invalid():
    """Test set_rist_regressor method with invalid estimator."""
    with pytest.raises(ValueError, match="UNKNOWN REGRESSOR"):
        _set_rist_regressor("invalid")



"""Tests for dataset resampling functions."""

__author__ = ["TonyBagnall", "MatthewMiddlehurst"]


import numpy as np
import pandas as pd
import pytest
from tsml.datasets import (
    load_equal_minimal_japanese_vowels,
    load_minimal_chinatown,
    load_unequal_minimal_chinatown,
)

from tsml_eval.testing.testing_utils import _TEST_RESULTS_PATH
from tsml_eval.utils.resampling import (
    resample_data,
    resample_data_indices,
    stratified_resample_data,
    stratified_resample_data_indices,
)
from tsml_eval.utils.results_validation import compare_result_file_resample


@pytest.mark.parametrize(
    "loader", [load_minimal_chinatown, load_equal_minimal_japanese_vowels]
)
def test_resample_data(loader):
    """Test resampling returns valid data."""
    X_train, y_train = loader(split="TRAIN")
    X_test, y_test = loader(split="TEST")

    train_size = X_train.shape
    test_size = X_test.shape

    X_train, y_train, X_test, y_test = resample_data(X_train, y_train, X_test, y_test)

    assert isinstance(X_train, np.ndarray)
    assert isinstance(X_test, np.ndarray)
    assert isinstance(y_test, np.ndarray)
    assert isinstance(y_test, np.ndarray)

    assert X_train.shape == train_size
    assert X_test.shape == test_size


def test_resample_data_unequal():
    """Test resampling returns valid data with unequal length input."""
    X_train, y_train = load_unequal_minimal_chinatown(split="TRAIN")
    X_test, y_test = load_unequal_minimal_chinatown(split="TEST")

    train_size = len(X_train)
    test_size = len(X_test)

    X_train, y_train, X_test, y_test = resample_data(X_train, y_train, X_test, y_test)

    assert isinstance(X_train, list)
    assert isinstance(X_train[0], np.ndarray)
    assert isinstance(X_test, list)
    assert isinstance(X_test[0], np.ndarray)
    assert isinstance(y_test, np.ndarray)
    assert isinstance(y_test, np.ndarray)

    assert len(X_train) == train_size
    assert len(X_test) == test_size


def test_resample_data_invalid():
    """Test resampling raises an error with invalid input."""
    X = pd.DataFrame(np.random.random((10, 10)))
    y = pd.Series(np.zeros(10))

    with pytest.raises(ValueError, match="X_train must be a"):
        resample_data(X, y, X, y)


def test_resample_data_indices():
    """Test resampling returns valid indices."""
    X_train, y_train = load_minimal_chinatown(split="TRAIN")
    X_test, y_test = load_minimal_chinatown(split="TEST")

    new_X_train, _, new_X_test, _ = resample_data(
        X_train, y_train, X_test, y_test, random_state=0
    )
    train_indices, test_indices = resample_data_indices(y_train, y_test, random_state=0)
    X = np.concatenate((X_train, X_test), axis=0)

    assert isinstance(train_indices, np.ndarray)
    assert isinstance(test_indices, np.ndarray)
    assert len(train_indices) == len(new_X_train)
    assert len(test_indices) == len(new_X_test)
    assert len(np.unique(np.concatenate((train_indices, test_indices), axis=0))) == len(
        X
    )
    assert (new_X_train[0] == X[train_indices[0]]).all()
    assert (new_X_test[0] == X[test_indices[0]]).all()


@pytest.mark.parametrize(
    "loader", [load_minimal_chinatown, load_equal_minimal_japanese_vowels]
)
def test_stratified_resample_data(loader):
    """Test stratified resampling returns valid data and class distribution."""
    X_train, y_train = loader(split="TRAIN")
    X_test, y_test = loader(split="TEST")

    train_size = X_train.shape
    test_size = X_test.shape
    _, counts_train = np.unique(y_train, return_counts=True)
    _, counts_test = np.unique(y_test, return_counts=True)

    X_train, y_train, X_test, y_test = stratified_resample_data(
        X_train, y_train, X_test, y_test
    )

    assert isinstance(X_train, np.ndarray)
    assert isinstance(X_test, np.ndarray)
    assert isinstance(y_test, np.ndarray)
    assert isinstance(y_test, np.ndarray)

    assert X_train.shape == train_size
    assert X_test.shape == test_size

    _, counts_train_new = np.unique(y_train, return_counts=True)
    _, counts_test_new = np.unique(y_test, return_counts=True)

    assert list(counts_train_new) == list(counts_train)
    assert list(counts_test_new) == list(counts_test)


def test_stratified_resample_data_unequal():
    """Test stratified resampling returns valid data with unequal length input."""
    X_train, y_train = load_unequal_minimal_chinatown(split="TRAIN")
    X_test, y_test = load_unequal_minimal_chinatown(split="TEST")

    train_size = len(X_train)
    test_size = len(X_test)
    _, counts_train = np.unique(y_train, return_counts=True)
    _, counts_test = np.unique(y_test, return_counts=True)

    X_train, y_train, X_test, y_test = stratified_resample_data(
        X_train, y_train, X_test, y_test
    )

    assert isinstance(X_train, list)
    assert isinstance(X_train[0], np.ndarray)
    assert isinstance(X_test, list)
    assert isinstance(X_test[0], np.ndarray)
    assert isinstance(y_test, np.ndarray)
    assert isinstance(y_test, np.ndarray)

    assert len(X_train) == train_size
    assert len(X_test) == test_size

    _, counts_train_new = np.unique(y_train, return_counts=True)
    _, counts_test_new = np.unique(y_test, return_counts=True)

    assert list(counts_train_new) == list(counts_train)
    assert list(counts_test_new) == list(counts_test)


def test_stratified_resample_data_invalid():
    """Test stratified resampling raises an error with invalid input."""
    X = pd.DataFrame(np.random.random((10, 10)))
    y = pd.Series(np.zeros(10))

    with pytest.raises(ValueError, match="X_train must be a"):
        stratified_resample_data(X, y, X, y)


def test_stratified_resample_data_indices():
    """Test stratified resampling returns valid indices."""
    X_train, y_train = load_minimal_chinatown(split="TRAIN")
    X_test, y_test = load_minimal_chinatown(split="TEST")

    new_X_train, _, new_X_test, _ = stratified_resample_data(
        X_train, y_train, X_test, y_test, random_state=0
    )
    train_indices, test_indices = stratified_resample_data_indices(
        y_train, y_test, random_state=0
    )
    X = np.concatenate((X_train, X_test), axis=0)

    assert isinstance(train_indices, np.ndarray)
    assert isinstance(test_indices, np.ndarray)
    assert len(train_indices) == len(new_X_train)
    assert len(test_indices) == len(new_X_test)
    assert len(np.unique(np.concatenate((train_indices, test_indices), axis=0))) == len(
        X
    )
    assert (new_X_train[0] == X[train_indices[0]]).all()
    assert (new_X_test[0] == X[test_indices[0]]).all()


@pytest.mark.parametrize(
    "paths",
    [
        [
            _TEST_RESULTS_PATH + "/classification/classificationResultsFile1.csv",
            _TEST_RESULTS_PATH + "/classification/classificationResultsFile1.csv",
            True,
        ],
        [
            _TEST_RESULTS_PATH + "/classification/classificationResultsFile1.csv",
            _TEST_RESULTS_PATH + "/classification/classificationResultsFile2.csv",
            False,
        ],
    ],
)
def test_compare_result_file_resample(paths):
    """Test compare result file resample function."""
    assert compare_result_file_resample(paths[0], paths[1]) == paths[2]


def test_compare_result_file_resample_invalid():
    """Test compare result file resample function with invalid input."""
    p1 = _TEST_RESULTS_PATH + "/classification/classificationResultsFile1.csv"
    p3 = _TEST_RESULTS_PATH + "/classification/classificationResultsFile3.csv"

    with pytest.raises(ValueError, match="Input results file have different"):
        compare_result_file_resample(p1, p3)



"""Class for storing and loading results from a forecasting experiment."""

import numpy as np
from sklearn.metrics import mean_absolute_percentage_error

from tsml_eval.evaluation.storage.estimator_results import EstimatorResults
from tsml_eval.utils.results_writing import write_forecasting_results


class ForecasterResults(EstimatorResults):
    """
    A class for storing and managing results from forecasting experiments.

    This class provides functionalities for storing forecaster results,
    including predictions, probabilities, and various performance metrics.
    It extends the `EstimatorResults` class, inheriting its base functionalities.

    Parameters
    ----------
    dataset_name : str, default="N/A"
        Name of the dataset used.
    forecaster_name : str, default="N/A"
        Name of the forecaster used.
    split : str, default="N/A"
        Type of data split used, i.e. "train" or "test".
    random_seed : int or None, default=None
        Random seed used.
    time_unit : str, default="nanoseconds"
        Time measurement used for other fields.
    description : str, default=""
        Additional description of the forecasting experiment. Appended to the end
        of the first line of the results file.
    parameters : str, default="No parameter info"
        Information about parameters used in the forecaster and other build information.
        Written to the second line of the results file.
    fit_time : float, default=-1.0
        Time taken fitting the model.
    predict_time : float, default=-1.0
        Time taken making predictions.
    benchmark_time : float, default=-1.0
        Time taken to run a simple benchmark function. In tsml-eval experiments, this
        is the time spent to sort 1,000 (seeded) random numpy arrays of size 20,000.
    memory_usage : float, default=-1.0
        Memory usage during the experiment. In tsml-eval experiments, this is the peak
        memory usage during the fit method.
    target_labels : array-like or None, default=None
        Actual target labels.
    predictions : array-like or None, default=None
        Predicted target labels.
    pred_times : array-like or None, default=None
        Prediction times for each case.
    pred_descriptions : list of str or None, default=None
        Descriptions for each prediction.

    Attributes
    ----------
    mean_absolute_percentage_error : float or None
        Mean absolute percentage error of the predictions.

    Examples
    --------
    >>> from tsml_eval.evaluation.storage import ForecasterResults
    >>> from tsml_eval.testing.testing_utils import _TEST_RESULTS_PATH
    >>> fr = ForecasterResults().load_from_file(
    ...     _TEST_RESULTS_PATH +
    ...     "/forecasting/NaiveForecaster/Predictions/Airline/testResample0.csv"
    ... )
    >>> fr.calculate_statistics()
    >>> fr.mean_absolute_percentage_error
    0.19886711926999853
    """

    def __init__(
        self,
        dataset_name="N/A",
        forecaster_name="N/A",
        split="N/A",
        random_seed=None,
        time_unit="nanoseconds",
        description="",
        parameters="No parameter info",
        fit_time=-1.0,
        predict_time=-1.0,
        benchmark_time=-1.0,
        memory_usage=-1.0,
        target_labels=None,
        predictions=None,
        pred_times=None,
        pred_descriptions=None,
    ):
        # Line 1
        self.forecaster_name = forecaster_name
        self.random_seed = random_seed

        # Results
        self.target_labels = target_labels
        self.predictions = predictions
        self.pred_times = pred_times
        self.pred_descriptions = pred_descriptions

        self.forecasting_horizon = None

        self.mean_absolute_percentage_error = None

        super().__init__(
            dataset_name=dataset_name,
            estimator_name=forecaster_name,
            split=split,
            resample_id=random_seed,
            time_unit=time_unit,
            description=description,
            parameters=parameters,
            fit_time=fit_time,
            predict_time=predict_time,
            benchmark_time=benchmark_time,
            memory_usage=memory_usage,
        )

    # var_name: (display_name, higher is better, is timing)
    statistics = {
        "mean_absolute_percentage_error": ("MAPE", False, False),
        **EstimatorResults.statistics,
    }

    def save_to_file(self, file_path, full_path=True):
        """
        Write the forecaster results into a file format used by tsml.

        Parameters
        ----------
        file_path : str
            Path to write the results file to or the directory to build the default file
            structure if full_path is False.
        full_path : boolean, default=True
            If True, results are written directly to the directory passed in file_path.
            If False, then a standard file structure using the forecaster and dataset
            names is created and used to write the results file.
        """
        self.infer_size()

        if self.mean_absolute_percentage_error is None:
            self.mean_absolute_percentage_error = mean_absolute_percentage_error(
                self.target_labels, self.predictions
            )

        write_forecasting_results(
            self.predictions,
            self.target_labels,
            self.estimator_name,
            self.dataset_name,
            file_path,
            full_path=full_path,
            split=self.split,
            random_seed=self.resample_id,
            time_unit=self.time_unit,
            first_line_comment=self.description,
            parameter_info=self.parameter_info,
            mape=self.mean_absolute_percentage_error,
            fit_time=self.fit_time,
            predict_time=self.predict_time,
            benchmark_time=self.benchmark_time,
            memory_usage=self.memory_usage,
        )

    def load_from_file(self, file_path, verify_values=True):
        """
        Load forecaster results from a specified file.

        This method reads a file containing forecaster results and reconstructs the
        ForecasterResults object. It calculates performance statistics and
        verifies values based on the loaded data.

        Parameters
        ----------
        file_path : str
            The path to the file from which forecaster results should be loaded. The
            file should be a tsml formatted forecaster results file.
        verify_values : bool, default=True
            If the method should perform verification of the loaded values.

        Returns
        -------
        self : ForecasterResults
            The same ForecasterResults object with loaded results.
        """
        fr = load_forecaster_results(file_path, verify_values=verify_values)
        self.__dict__.update(fr.__dict__)
        return self

    def calculate_statistics(self, overwrite=False):
        """
        Calculate various performance statistics based on the forecaster results.

        This method computes various performance metrics, such as MAPE based on the
        forecasters output.

        Parameters
        ----------
        overwrite : bool, default=False
            If the function should overwrite the current values when they are not None.
        """
        self.infer_size(overwrite=overwrite)

        if self.mean_absolute_percentage_error is None or overwrite:
            self.mean_absolute_percentage_error = mean_absolute_percentage_error(
                self.target_labels, self.predictions
            )

    def infer_size(self, overwrite=False):
        """
        Infer and return the size of the dataset used in the results.

        This method estimates the size of the dataset that was used for the estimator,
        based on the results data.

        Parameters
        ----------
        overwrite : bool, default=False
            If the function should overwrite the current values when they are not None.
        """
        if self.forecasting_horizon is None or overwrite:
            self.forecasting_horizon = len(self.target_labels)


def load_forecaster_results(file_path, calculate_stats=True, verify_values=True):
    """
    Load and return forecaster results from a specified file.

    This function reads a file containing forecaster results and reconstructs the
    ForecasterResults object. It optionally calculates performance statistics and
    verifies values based on the loaded data.

    Parameters
    ----------
    file_path : str
        The path to the file from which forecaster results should be loaded. The file
        should be a tsml formatted forecaster results file.
    calculate_stats : bool, default=True
        Whether to calculate performance statistics from the loaded results.
    verify_values : bool, default=True
        If the function should perform verification of the loaded values.

    Returns
    -------
    fr : ForecasterResults
        A ForecasterResults object containing the results loaded from the file.
    """
    with open(file_path) as file:
        lines = file.readlines()

        line1 = lines[0].split(",")
        line3 = lines[2].split(",")
        mape = float(line3[0])
        fh = len(lines) - 3

        line_size = len(lines[3].split(","))

        target_labels = np.zeros(fh)
        predictions = np.zeros(fh)

        if line_size > 3:
            pred_times = np.zeros(fh)
        else:
            pred_times = None

        if line_size > 5:
            pred_descriptions = []
        else:
            pred_descriptions = None

        for i in range(0, fh):
            line = lines[i + 3].split(",")
            target_labels[i] = float(line[0])
            predictions[i] = float(line[1])

            if pred_times is not None:
                pred_times[i] = float(line[3])

            if pred_descriptions is not None:
                pred_descriptions.append(",".join(line[5]).strip())

    fr = ForecasterResults(
        dataset_name=line1[0],
        forecaster_name=line1[1],
        split=line1[2],
        random_seed=None if line1[3] == "None" else int(line1[3]),
        time_unit=line1[4].lower(),
        description=",".join(line1[5:]).strip(),
        parameters=lines[1].strip(),
        fit_time=float(line3[1]),
        predict_time=float(line3[2]),
        benchmark_time=float(line3[3]),
        memory_usage=float(line3[4]),
        target_labels=target_labels,
        predictions=predictions,
        pred_times=pred_times,
        pred_descriptions=pred_descriptions,
    )

    if calculate_stats:
        fr.calculate_statistics()

    if verify_values:
        fr.infer_size(overwrite=True)
        assert fr.forecasting_horizon == fh

        if calculate_stats:
            assert fr.mean_absolute_percentage_error == mape

    return fr



"""Utilities for publications."""

__all__ = [
    "extract_publication_csv_from_evaluation",
    "parameter_table_from_estimator_selector",
]


import os
import shutil


def extract_publication_csv_from_evaluation(stats, eval_path, write_path):
    """Extract the CSV files from the evaluation directory to a new directory.

    Parameters
    ----------
    stats : str or list of str
        The statistics to extract. The name of a statistic or list of names.
    eval_path : str or list of str
        The path to the evaluation directory generated from the evaluation package.
        If a list, each item is the path to an evaluation directory.
    write_path : str or list of str
        The path to the directory where the CSV files will be written. If a list,
        each item is the path to a directory where the CSV files will be written.
        eval_path and write_path must be the same length.

    Examples
    --------
    >>> from tsml_eval.testing.testing_utils import (
    ...     _TEST_EVAL_PATH, _TEST_OUTPUT_PATH
    ... )
    >>> from tsml_eval.utils.publications import (
    ...     extract_publication_csv_from_evaluation
    ... )
    >>> extract_publication_csv_from_evaluation(
    ...     ["Accuracy", "BalAcc"],
    ...     f"{_TEST_EVAL_PATH}/classification/",
    ...     f"{_TEST_OUTPUT_PATH}/result_files/",
    ... )
    """
    if isinstance(stats, str):
        stats = [stats]

    os.makedirs(write_path, exist_ok=True)

    for stat in stats:
        stat_dir = f"{eval_path}/{stat}/all_resamples/"

        for file in os.listdir(stat_dir):
            if file.endswith(".csv"):
                shutil.copy(f"{stat_dir}/{file}", f"{write_path}/{file}")


def parameter_table_from_estimator_selector(selection_function, estimator_names):
    """Create a table of estimator names and their parameters.

    Parameters
    ----------
    selection_function : function
        The function that selects the estimator.
    estimator_names : list of str
        The names of the estimators.
    """
    parameters = []
    for estimator_name in estimator_names:
        est = selection_function(estimator_name)
        params = est.get_params()
        params.pop("random_state", None)
        params.pop("n_jobs", None)
        params.pop("verbose", None)
        parameters.append(params)

    table = "Estimator & Parameters \\\\ \\hline \n"
    for i, params in enumerate(parameters):
        table += estimator_names[i] + " & "
        for key, value in params.items():
            table += f"{key}: {value}, "
        table += " \\\\ \n"

    return table



"""Test file checking that the TSC bakeoff experiments run correctly."""

import os
import runpy

from tsml_eval.publications.y2023.tsc_bakeoff.run_experiments import _run_experiment
from tsml_eval.publications.y2023.tsc_bakeoff.tests import _BAKEOFF_TEST_RESULTS_PATH
from tsml_eval.testing.testing_utils import _TEST_DATA_PATH
from tsml_eval.utils.tests.test_results_writing import _check_classification_file_format


def test_run_tsc_bakeoff_experiment():
    """Test paper classification experiments with test data and classifier."""
    classifier = "ROCKET"
    dataset = "MinimalChinatown"
    resample = 0

    args = [
        _TEST_DATA_PATH,
        _BAKEOFF_TEST_RESULTS_PATH,
        classifier,
        dataset,
        resample,
        "-ow",
    ]

    _run_experiment(args, predefined_resample=False)

    test_file = (
        f"{_BAKEOFF_TEST_RESULTS_PATH}{classifier}/Predictions/{dataset}/"
        "testResample0.csv"
    )
    assert os.path.exists(test_file)
    _check_classification_file_format(test_file)

    # this covers both the main method and present result file checking
    runpy.run_path(
        (
            "./tsml_eval/publications/y2023/tsc_bakeoff/run_experiments.py"
            if os.getcwd().split("\\")[-1] != "tests"
            else "../run_experiments.py"
        ),
        run_name="__main__",
    )

    os.remove(test_file)



"""TSER expansion publication tests."""

__all__ = ["_TSER_ARCHIVE_TEST_RESULTS_PATH"]

import os
from pathlib import Path

_TSER_ARCHIVE_TEST_RESULTS_PATH = (
    os.path.dirname(Path(__file__).parent.parent.parent.parent.parent)
    + "/test_output/expansion_regression/"
)



"""Tests for results writing functions."""

__author__ = ["MatthewMiddlehurst"]

import os

import numpy as np
import pytest

from tsml_eval.experiments.tests import (
    _CLASSIFIER_RESULTS_PATH,
    _CLUSTERER_RESULTS_PATH,
    _FORECASTER_RESULTS_PATH,
    _REGRESSOR_RESULTS_PATH,
)
from tsml_eval.utils.results_validation import (
    _check_classification_third_line,
    _check_clustering_third_line,
    _check_first_line,
    _check_forecasting_third_line,
    _check_regression_third_line,
    _check_results_lines,
    _check_second_line,
)
from tsml_eval.utils.results_writing import (
    write_classification_results,
    write_clustering_results,
    write_forecasting_results,
    write_regression_results,
    write_results_to_tsml_format,
)


def test_write_classification_results():
    """Test writing of classification results files."""
    class_labels, predictions, probabilities = _generate_labels_and_predictions()

    write_classification_results(
        predictions,
        probabilities,
        class_labels,
        "Test",
        "Test",
        _CLASSIFIER_RESULTS_PATH,
        full_path=False,
        first_line_comment="test_write_classification_results",
        n_classes=3,
    )

    _check_classification_file_format(
        f"{_CLASSIFIER_RESULTS_PATH}/Test/Predictions/Test/results.csv"
    )

    os.remove(f"{_CLASSIFIER_RESULTS_PATH}/Test/Predictions/Test/results.csv")


def _check_classification_file_format(file_path, num_results_lines=None):
    with open(file_path) as f:
        lines = f.readlines()

    assert _check_first_line(lines[0])
    assert _check_second_line(lines[1])
    assert _check_classification_third_line(lines[2])
    n_classes = int(lines[2].split(",")[5])

    _check_results_lines(lines, num_results_lines=num_results_lines, n_probas=n_classes)


def test_write_classification_results_invalid():
    """Test writing of classification results files with invalid input."""
    with pytest.raises(IndexError, match="The number of predicted values"):
        write_classification_results(
            np.zeros(10),
            np.zeros((11, 2)),
            np.zeros(12),
            "Test",
            "Test",
            "test_output",
        )

    with pytest.raises(IndexError, match="The number of classes is not"):
        write_classification_results(
            np.zeros(10),
            np.zeros((10, 3)),
            np.zeros(10),
            "Test",
            "Test",
            "test_output",
            n_classes=2,
        )


def test_write_regression_results():
    """Test writing of regression results files."""
    labels, predictions, _ = _generate_labels_and_predictions()

    write_regression_results(
        predictions,
        labels,
        "Test",
        "Test",
        _REGRESSOR_RESULTS_PATH,
        full_path=False,
        first_line_comment="test_write_regression_results",
    )

    _check_regression_file_format(
        f"{_REGRESSOR_RESULTS_PATH}/Test/Predictions/Test/results.csv"
    )

    os.remove(f"{_REGRESSOR_RESULTS_PATH}/Test/Predictions/Test/results.csv")


def _check_regression_file_format(file_path, num_results_lines=None):
    with open(file_path) as f:
        lines = f.readlines()

    assert _check_first_line(lines[0])
    assert _check_second_line(lines[1])
    assert _check_regression_third_line(lines[2])

    _check_results_lines(
        lines, num_results_lines=num_results_lines, probabilities=False
    )


def test_write_forecasting_results():
    """Test writing of forecasting results files."""
    labels, predictions, _ = _generate_labels_and_predictions()

    write_forecasting_results(
        predictions,
        labels,
        "Test",
        "Test",
        _FORECASTER_RESULTS_PATH,
        full_path=False,
        first_line_comment="test_write_forecasting_results",
    )

    _check_forecasting_file_format(
        f"{_FORECASTER_RESULTS_PATH}/Test/Predictions/Test/results.csv"
    )

    os.remove(f"{_FORECASTER_RESULTS_PATH}/Test/Predictions/Test/results.csv")


def _check_forecasting_file_format(file_path, num_results_lines=None):
    with open(file_path) as f:
        lines = f.readlines()

    assert _check_first_line(lines[0])
    assert _check_second_line(lines[1])
    assert _check_forecasting_third_line(lines[2])

    _check_results_lines(
        lines, num_results_lines=num_results_lines, probabilities=False
    )


def test_write_clustering_results():
    """Test writing of clustering results files."""
    (
        class_labels,
        cluster_predictions,
        cluster_probabilities,
    ) = _generate_labels_and_predictions()

    write_clustering_results(
        cluster_predictions,
        cluster_probabilities,
        class_labels,
        "Test",
        "Test",
        _CLUSTERER_RESULTS_PATH,
        full_path=False,
        first_line_comment="test_write_clustering_results",
        n_clusters=3,
    )

    _check_clustering_file_format(
        f"{_CLUSTERER_RESULTS_PATH}/Test/Predictions/Test/results.csv"
    )

    os.remove(f"{_CLUSTERER_RESULTS_PATH}/Test/Predictions/Test/results.csv")


def _check_clustering_file_format(file_path, num_results_lines=None):
    with open(file_path) as f:
        lines = f.readlines()

    assert _check_first_line(lines[0])
    assert _check_second_line(lines[1])
    assert _check_clustering_third_line(lines[2])
    n_probas = int(lines[2].split(",")[6])

    _check_results_lines(lines, num_results_lines=num_results_lines, n_probas=n_probas)


def test_write_clustering_results_invalid():
    """Test writing of clustering results files with invalid input."""
    with pytest.raises(IndexError, match="The number of predicted values"):
        write_clustering_results(
            np.zeros(10),
            np.zeros((11, 2)),
            np.zeros(12),
            "Test",
            "Test",
            "test_output",
        )

    with pytest.raises(IndexError, match="The number of clusters is not"):
        write_clustering_results(
            np.zeros(10),
            np.zeros((10, 3)),
            np.zeros(10),
            "Test",
            "Test",
            "test_output",
            n_clusters=2,
        )


def _generate_labels_and_predictions():
    labels = np.random.randint(0, 2, 10)
    predictions = np.random.randint(0, 2, 10)

    probabilities = np.zeros((10, 3))
    for i in range(10):
        probabilities[i, predictions[i]] = 1

    return labels, predictions, probabilities


def test_write_results_to_tsml_format_invalid():
    """Test writing of results files with invalid input."""
    with pytest.raises(IndexError, match="The number of predicted values"):
        write_results_to_tsml_format(
            np.zeros(10),
            np.zeros(11),
            "Test",
            "Test",
            "test_output",
        )

    with pytest.raises(ValueError, match="Unknown 'split' value"):
        write_results_to_tsml_format(
            np.zeros(10),
            np.zeros(10),
            "Test",
            "Test",
            "test_output",
            split="invalid",
        )



"""Utility functions for repairing results."""

__author__ = ["MatthewMiddlehurst"]
__all__ = ["fix_broken_second_line"]

import os

from tsml_eval.utils.results_validation import (
    _check_classification_third_line,
    _check_clustering_third_line,
    _check_forecasting_third_line,
    _check_regression_third_line,
)


def fix_broken_second_line(file_path, save_path=None):
    """Fix a results while where the written second line has line breaks.

    This function will remove line breaks from any lines between the first line and the
    first seen valid 'third_line' for any results file format.

    Parameters
    ----------
    file_path : str
        Path to the results file to be fixed, including the file itself.
    save_path : str, default=None
        Path to save the fixed results file to, including the file new files name.
        If None, the new file will replace the original file.
    """
    with open(file_path) as f:
        lines = f.readlines()

    line_count = 2
    while (
        not _check_classification_third_line(lines[line_count])
        and not _check_regression_third_line(lines[line_count])
        and not _check_clustering_third_line(lines[line_count])
        and not _check_forecasting_third_line(lines[line_count])
    ):
        if line_count == len(lines) - 1:
            raise ValueError("No valid third line found in input results file.")
        line_count += 1

    if line_count != 2:
        lines[1] = lines[1].replace("\n", " ").replace("\r", " ")
        for i in range(2, line_count - 1):
            lines[1] = lines[1] + lines[i].replace("\n", " ").replace("\r", " ")
        lines[1] = lines[1] + lines[line_count - 1]
        lines = lines[:2] + lines[line_count:]

    if save_path is not None or line_count != 2:
        if save_path is None:
            save_path = file_path

        os.makedirs(os.path.dirname(save_path), exist_ok=True)

        with open(save_path, "w") as f:
            f.writelines(lines)



"""Tests for the multiple estimator evaluation functionality."""

from tsml_eval.evaluation.multiple_estimator_evaluation import (
    evaluate_classifiers_by_problem,
    evaluate_clusterers_by_problem,
    evaluate_forecasters_by_problem,
    evaluate_regressors_by_problem,
)
from tsml_eval.testing.testing_utils import _TEST_OUTPUT_PATH, _TEST_RESULTS_PATH


def test_evaluate_classifiers_by_problem():
    """Test the evaluation of classifiers by problem."""
    classifiers = ["ROCKET", "TSF", "1NN-DTW"]
    datasets = ["Chinatown", "ItalyPowerDemand", "Trace"]
    resamples = 3

    evaluate_classifiers_by_problem(
        _TEST_RESULTS_PATH + "/classification/",
        classifiers,
        datasets,
        _TEST_OUTPUT_PATH + "/eval/classification/",
        resamples=resamples,
        eval_name="test0",
    )


def test_evaluate_clusterers_by_problem():
    """Test the evaluation of clusterers by problem."""
    classifiers = ["KMeans", "KMeans-dtw", "KMeans-msm"]
    datasets = ["Chinatown", "ItalyPowerDemand", "Trace"]
    resamples = 3

    evaluate_clusterers_by_problem(
        _TEST_RESULTS_PATH + "/clustering/",
        classifiers,
        datasets,
        _TEST_OUTPUT_PATH + "/eval/clustering/",
        resamples=resamples,
        load_test_results=False,
        eval_name="test0",
    )


def test_evaluate_regressors_by_problem():
    """Test the evaluation of regressors by problem."""
    classifiers = ["ROCKET", "TSF", "1NN-DTW"]
    datasets = ["Covid3Month", "NaturalGasPricesSentiment", "FloodModeling1"]
    resamples = 3

    evaluate_regressors_by_problem(
        _TEST_RESULTS_PATH + "/regression/",
        classifiers,
        datasets,
        _TEST_OUTPUT_PATH + "/eval/regression/",
        resamples=resamples,
        eval_name="test0",
    )


def test_evaluate_forecasters_by_problem():
    """Test the evaluation of forecasters by problem."""
    classifiers = ["NaiveForecaster", "RandomForest", "LinearRegression"]
    datasets = ["Airline", "ShampooSales"]
    resamples = 1

    evaluate_forecasters_by_problem(
        _TEST_RESULTS_PATH + "/forecasting/",
        classifiers,
        datasets,
        _TEST_OUTPUT_PATH + "/eval/forecasting/",
        resamples=resamples,
        eval_name="test0",
    )



"""Test the SimpleVote."""

import numpy as np
from aeon.datasets import load_arrow_head
from sklearn.metrics import rand_score

from tsml_eval.estimators.clustering.consensus.simple_vote_from_file import (
    FromFileSimpleVote,
)
from tsml_eval.testing.testing_utils import _TEST_RESULTS_PATH


def test_from_file_simple_vote():
    """Test SimpleVote from file with ArrowHead results."""
    X_train, y_train = load_arrow_head(split="train")
    X_test, y_test = load_arrow_head(split="test")

    file_paths = [
        _TEST_RESULTS_PATH + "/clustering/PAM-DTW/Predictions/ArrowHead/",
        _TEST_RESULTS_PATH + "/clustering/PAM-ERP/Predictions/ArrowHead/",
        _TEST_RESULTS_PATH + "/clustering/PAM-MSM/Predictions/ArrowHead/",
    ]

    sv = FromFileSimpleVote(clusterers=file_paths, n_clusters=3, random_state=0)
    sv.fit(X_train, y_train)
    preds = sv.predict(X_test)

    assert sv.labels_.shape == (len(X_train),)
    assert isinstance(sv.labels_, np.ndarray)
    assert rand_score(y_train, sv.labels_) >= 0.6
    assert preds.shape == (len(X_test),)
    assert isinstance(preds, np.ndarray)
    assert rand_score(y_test, preds) >= 0.6



"""Test experiment utilities."""

import os

import pytest
from aeon.classification.shapelet_based import ShapeletTransformClassifier
from aeon.classification.sklearn import RotationForestClassifier
from tsml.datasets import load_minimal_chinatown

from tsml_eval.testing.testing_utils import _TEST_OUTPUT_PATH
from tsml_eval.utils.experiments import (
    _results_present,
    estimator_attributes_to_file,
    timing_benchmark,
)


@pytest.mark.parametrize("split", ["BOTH", "TRAIN", "TEST", None, "invalid"])
def test_results_present_split_inputs(split):
    """Test _results_present function with valid and invalid split inputs."""
    if split == "invalid":
        with pytest.raises(ValueError, match="Unknown split value"):
            _results_present(
                "test_output",
                "test",
                "test",
                split=split,
            )
    else:
        assert not _results_present(
            "test_output",
            "test",
            "test",
            split=split,
        )


def test_timing_benchmark_invalid_input():
    """Test timing_benchmark function with invalid input."""
    with pytest.raises(ValueError):
        timing_benchmark(random_state="invalid")


def test_estimator_attributes_to_file():
    """Test writing estimator attributes to file."""
    estimator = ShapeletTransformClassifier(
        n_shapelet_samples=50,
        estimator=RotationForestClassifier(n_estimators=2),
    )
    X, y = load_minimal_chinatown()
    estimator.fit(X, y)

    test_dir = _TEST_OUTPUT_PATH + "/attribute_writing/"
    estimator_attributes_to_file(estimator, test_dir)

    assert os.path.exists(test_dir + "ShapeletTransformClassifier.txt")
    assert os.path.exists(test_dir + "estimator/estimator.txt")
    assert os.path.exists(test_dir + "_estimator/_estimator.txt")
    assert os.path.exists(test_dir + "_estimator/_base_estimator/_base_estimator.txt")
    assert os.path.exists(test_dir + "_estimator/_pcas_0_0/_pcas_0_0.txt")


def test_max_depth():
    """Test writing estimator attributes to file max depth parameter."""
    estimator = ShapeletTransformClassifier(
        n_shapelet_samples=50,
        estimator=RotationForestClassifier(n_estimators=2),
    )
    X, y = load_minimal_chinatown()
    estimator.fit(X, y)

    test_dir = _TEST_OUTPUT_PATH + "/attribute_writing_max_depth/"
    estimator_attributes_to_file(estimator, test_dir, max_depth=1)

    assert os.path.exists(test_dir + "ShapeletTransformClassifier.txt")
    assert os.path.exists(test_dir + "estimator/estimator.txt")
    assert os.path.exists(test_dir + "_estimator/_estimator.txt")
    assert not os.path.exists(
        test_dir + "_estimator/_base_estimator/_base_estimator.txt"
    )
    assert not os.path.exists(test_dir + "_estimator/_pcas_0_0/_pcas_0_0.txt")


def test_max_list_shape():
    """Test writing estimator attributes to file max list shape parameter."""
    estimator = ShapeletTransformClassifier(
        n_shapelet_samples=50,
        estimator=RotationForestClassifier(n_estimators=2),
    )
    X, y = load_minimal_chinatown()
    estimator.fit(X, y)

    test_dir = _TEST_OUTPUT_PATH + "/attribute_writing_max_list_shape/"
    estimator_attributes_to_file(estimator, test_dir, max_list_shape=1)

    assert os.path.exists(test_dir + "ShapeletTransformClassifier.txt")
    assert os.path.exists(test_dir + "estimator/estimator.txt")
    assert os.path.exists(test_dir + "_estimator/_estimator.txt")
    assert os.path.exists(test_dir + "_estimator/_base_estimator/_base_estimator.txt")
    assert not os.path.exists(test_dir + "_estimator/_pcas_0_0/_pcas_0_0.txt")



"""Functions for evaluating multiple estimators on multiple datasets."""

import os
import pickle
import warnings
from datetime import datetime

import numpy as np
from aeon.performance_metrics.stats import wilcoxon_test
from aeon.visualisation import (
    plot_boxplot_median,
    plot_critical_difference,
    plot_pairwise_scatter,
)
from matplotlib import pyplot as plt

from tsml_eval.evaluation.storage import (
    ClassifierResults,
    ClustererResults,
    ForecasterResults,
    RegressorResults,
)
from tsml_eval.utils.functions import rank_array, time_to_milliseconds

__all__ = [
    "evaluate_classifiers",
    "evaluate_classifiers_from_file",
    "evaluate_classifiers_by_problem",
    "evaluate_clusterers",
    "evaluate_clusterers_from_file",
    "evaluate_clusterers_by_problem",
    "evaluate_regressors",
    "evaluate_regressors_from_file",
    "evaluate_regressors_by_problem",
    "evaluate_forecasters",
    "evaluate_forecasters_from_file",
    "evaluate_forecasters_by_problem",
]


def evaluate_classifiers(
    classifier_results,
    save_path,
    error_on_missing=True,
    eval_name=None,
    estimator_names=None,
):
    """
    Evaluate multiple classifiers on multiple datasets.

    Writes multiple csv files and figures to save_path, one for each statistic
    evaluated. Provides a summary csv file with the average statistic and
    average rank for each classifier.

    Parameters
    ----------
    classifier_results : list of ClassifierResults
        The results to evaluate.
    save_path : str
        The path to save the evaluation results to.
    error_on_missing : bool, default=True
        Whether to raise an error if results are missing.
    eval_name : str, default=None
        The name of the evaluation, used in save_path.
    estimator_names : list of str, default=None
        The names of the estimator for each classifier result. If None, uses
        the estimator_name attribute of each classifier result.
    """
    _evaluate_estimators(
        classifier_results,
        ClassifierResults.statistics,
        save_path,
        error_on_missing,
        eval_name,
        estimator_names,
    )


def evaluate_classifiers_from_file(
    load_paths,
    save_path,
    error_on_missing=True,
    eval_name=None,
    verify_results=True,
    estimator_names=None,
):
    """
    Evaluate multiple classifiers on multiple datasets from file.

    Writes multiple csv files and figures to save_path, one for each statistic
    evaluated. Provides a summary csv file with the average statistic and
    average rank for each classifier.

    Parameters
    ----------
    load_paths : list of str
        The paths to the classifier result files to evaluate.
    save_path : str
        The path to save the evaluation results to.
    error_on_missing : bool, default=True
        Whether to raise an error if results are missing.
    eval_name : str, default=None
        The name of the evaluation, used in save_path.
    verify_results : bool, default=True
        If the verification should be performed on the loaded results values.
    estimator_names : list of str, default=None
        The names of the estimator for each classifier result. If None, uses
        the estimator_name attribute of each classifier result.
    """
    classifier_results = []
    for load_path in load_paths:
        try:
            classifier_results.append(
                ClassifierResults().load_from_file(
                    load_path, verify_values=verify_results
                )
            )
        except FileNotFoundError:
            if error_on_missing:
                raise FileNotFoundError(f"Results for {load_path} not found.")

    evaluate_classifiers(
        classifier_results,
        save_path,
        error_on_missing=error_on_missing,
        eval_name=eval_name,
        estimator_names=estimator_names,
    )


def evaluate_classifiers_by_problem(
    load_path,
    classifier_names,
    dataset_names,
    save_path,
    resamples=None,
    load_train_results=False,
    error_on_missing=True,
    eval_name=None,
    verify_results=True,
    verbose=False,
):
    """
    Evaluate multiple classifiers on multiple datasets from file using standard paths.

    Finds files using classifier, dataset and resample names. It is expected the
    common tsml-eval file structure of
    {classifier}/Predictions/{dataset}/{split}Resample{resample}.csv is followed.

    Writes multiple csv files and figures to save_path, one for each statistic
    evaluated. Provides a summary csv file with the average statistic and
    average rank for each classifier.

    Parameters
    ----------
    load_path : str or list of str
        The path to the collection of classifier result files to evaluate.
        If load_path is a list, it will load results from each path in the list. It
        is expected that classifier_names and dataset_names are lists of lists with
        the same length as load_path.
    classifier_names : list of str, list of tuple or list of list
        The names of the classifiers to evaluate.
        A length 2 tuple containing strings can be used to specify a classifier name to
        load from in the first item and a classifier name to use in the evaluation
        results in the second.
        If load_path is a list, classifier_names must be a list of lists with the same
        length as load_path.
    dataset_names : str, list of str or list of list
        The names of the datasets to evaluate. If a list of strings, each item is the
        name of a dataset. If a string, it is the path to a file containing the names
        of the datasets, one per line.
        If load_path is a list, dataset_names must be a list of lists with the same
        length as load_path.
    save_path : str
        The path to save the evaluation results to.
    resamples : int or list of int, default=None
        The resamples to evaluate. If int, evaluates resamples 0 to resamples-1.
    load_train_results : bool, default=False
        Whether to load train results as well as test results.
    error_on_missing : bool, default=True
        Whether to raise an error if results are missing.
    eval_name : str, default=None
        The name of the evaluation, used in save_path.
    verify_results : bool, default=True
        If the verification should be performed on the loaded results values.
    verbose : bool, default=False
        If verbose output should be printed.
    """
    load_path, classifier_names, dataset_names, resamples = _evaluate_by_problem_init(
        "classifier",
        load_path,
        classifier_names,
        dataset_names,
        resamples,
    )

    if load_train_results:
        splits = ["test", "train"]
    else:
        splits = ["test"]

    classifier_results = []
    estimator_eval_names = []
    names = []
    for i, path in enumerate(load_path):
        found_datasets = np.zeros(len(dataset_names[i]), dtype=bool)

        for classifier_name in classifier_names[i]:
            found_estimator = False

            if isinstance(classifier_name, tuple):
                classifier_eval_name = classifier_name[1]
                classifier_name = classifier_name[0]
            else:
                classifier_eval_name = classifier_name

            if classifier_eval_name not in estimator_eval_names:
                estimator_eval_names.append(classifier_eval_name)
            else:
                raise ValueError(
                    f"Duplicate evaluation name {classifier_eval_name} found."
                )

            for n, dataset_name in enumerate(dataset_names[i]):
                for resample in resamples:
                    for split in splits:
                        try:
                            result = ClassifierResults().load_from_file(
                                f"{path}/{classifier_name}/Predictions/"
                                f"{dataset_name}/{split}Resample{resample}.csv",
                                verify_values=verify_results,
                            )
                            classifier_results.append(result)
                            names.append(classifier_eval_name)
                            found_estimator = True
                            found_datasets[n] = True

                            if verbose:
                                print(  # noqa: T201
                                    f"Loaded {classifier_eval_name} on {dataset_name} "
                                    f"{split} resample {resample}."
                                )

                            if result.dataset_name != dataset_name:
                                if verbose:
                                    print(  # noqa: T201
                                        f"File for {classifier_eval_name} on "
                                        f"{dataset_name} {split} resample {resample} "
                                        f"has a different dataset name than used to "
                                        f"load the file. Setting dataset name to "
                                        f"{dataset_name} from {result.dataset_name}."
                                    )
                                result.dataset_name = dataset_name
                        except FileNotFoundError:
                            msg = (
                                f"Results for {classifier_eval_name} on {dataset_name} "
                                f"{split} resample {resample} not found."
                            )
                            if error_on_missing:
                                raise FileNotFoundError(msg)
                            elif verbose:
                                print(msg)  # noqa: T201

            if not found_estimator:
                print(f"Classifier {classifier_eval_name} not found.")  # noqa: T201

        missing_datasets = [
            dataset
            for dataset, found in zip(dataset_names[i], found_datasets)
            if not found
        ]
        if missing_datasets:
            msg = f"Files for datasets {missing_datasets} not found."
            if error_on_missing:
                raise FileNotFoundError(msg)
            else:
                print("\n\n" + msg)  # noqa: T201

    evaluate_classifiers(
        classifier_results,
        save_path,
        error_on_missing=error_on_missing,
        eval_name=eval_name,
        estimator_names=names,
    )


def evaluate_clusterers(
    clusterer_results,
    save_path,
    error_on_missing=True,
    eval_name=None,
    estimator_names=None,
):
    """
    Evaluate multiple clusterers on multiple datasets.

    Writes multiple csv files and figures to save_path, one for each statistic
    evaluated. Provides a summary csv file with the average statistic and
    average rank for each clusterer.

    Parameters
    ----------
    clusterer_results : list of ClustererResults
        The results to evaluate.
    save_path : str
        The path to save the evaluation results to.
    error_on_missing : bool, default=True
        Whether to raise an error if results are missing.
    eval_name : str, default=None
        The name of the evaluation, used in save_path.
    estimator_names : list of str, default=None
        The names of the estimator for each clusterer result. If None, uses
        the estimator_name attribute of each clusterer result.
    """
    _evaluate_estimators(
        clusterer_results,
        ClustererResults.statistics,
        save_path,
        error_on_missing,
        eval_name,
        estimator_names,
    )


def evaluate_clusterers_from_file(
    load_paths,
    save_path,
    error_on_missing=True,
    eval_name=None,
    verify_results=True,
    estimator_names=None,
):
    """
    Evaluate multiple clusterers on multiple datasets from file.

    Writes multiple csv files and figures to save_path, one for each statistic
    evaluated. Provides a summary csv file with the average statistic and
    average rank for each clusterer.

    Parameters
    ----------
    load_paths : list of str
        The paths to the clusterer result files to evaluate.
    save_path : str
        The path to save the evaluation results to.
    error_on_missing : bool, default=True
        Whether to raise an error if results are missing.
    eval_name : str, default=None
        The name of the evaluation, used in save_path.
    verify_results : bool, default=True
        If the verification should be performed on the loaded results values.
    estimator_names : list of str, default=None
        The names of the estimator for each clusterer result. If None, uses
        the estimator_name attribute of each clusterer result.
    """
    clusterer_results = []
    for load_path in load_paths:
        try:
            clusterer_results.append(
                ClustererResults().load_from_file(
                    load_path, verify_values=verify_results
                )
            )
        except FileNotFoundError:
            if error_on_missing:
                raise FileNotFoundError(f"Results for {load_path} not found.")

    evaluate_clusterers(
        clusterer_results,
        save_path,
        error_on_missing=error_on_missing,
        eval_name=eval_name,
        estimator_names=estimator_names,
    )


def evaluate_clusterers_by_problem(
    load_path,
    clusterer_names,
    dataset_names,
    save_path,
    resamples=None,
    load_test_results=True,
    error_on_missing=True,
    eval_name=None,
    verify_results=True,
    verbose=False,
):
    """
    Evaluate multiple clusterers on multiple datasets from file using standard paths.

    Finds files using clusterer, dataset and resample names. It is expected the
    common tsml-eval file structure of
    {clusterer}/Predictions/{dataset}/{split}Resample{resample}.csv is followed.

    Writes multiple csv files and figures to save_path, one for each statistic
    evaluated. Provides a summary csv file with the average statistic and
    average rank for each clusterer.

    Parameters
    ----------
    load_path : str or list of str
        The path to the collection of clusterer result files to evaluate.
        If load_path is a list, it will load results from each path in the list. It
        is expected that clusterer_names and dataset_names are lists of lists with
        the same length as load_path.
    clusterer_names : list of str, list of tuple or list of list
        The names of the clusterers to evaluate.
        A length 2 tuple containing strings can be used to specify a clusterer name to
        load from in the first item and a clusterer name to use in the evaluation
        results in the second.
        If load_path is a list, clusterer_names must be a list of lists with the same
        length as load_path.
    dataset_names : str, list of str or list of list
        The names of the datasets to evaluate. If a list of strings, each item is the
        name of a dataset. If a string, it is the path to a file containing the names
        of the datasets, one per line.
        If load_path is a list, dataset_names must be a list of lists with the same
        length as load_path.
    save_path : str
        The path to save the evaluation results to.
    resamples : int or list of int, default=None
        The resamples to evaluate. If int, evaluates resamples 0 to resamples-1.
    load_test_results : bool, default=True
        Whether to load test results as well as train results.
    error_on_missing : bool, default=True
        Whether to raise an error if results are missing.
    eval_name : str, default=None
        The name of the evaluation, used in save_path.
    verify_results : bool, default=True
        If the verification should be performed on the loaded results values.
    verbose : bool, default=False
        If verbose output should be printed.
    """
    load_path, clusterer_names, dataset_names, resamples = _evaluate_by_problem_init(
        "clusterer",
        load_path,
        clusterer_names,
        dataset_names,
        resamples,
    )

    if load_test_results:
        splits = ["test", "train"]
    else:
        splits = ["train"]

    clusterer_results = []
    estimator_eval_names = []
    names = []
    for i, path in enumerate(load_path):
        found_datasets = np.zeros(len(dataset_names[i]), dtype=bool)

        for clusterer_name in clusterer_names[i]:
            found_estimator = False

            if isinstance(clusterer_name, tuple):
                clusterer_eval_name = clusterer_name[1]
                clusterer_name = clusterer_name[0]
            else:
                clusterer_eval_name = clusterer_name

            if clusterer_eval_name not in estimator_eval_names:
                estimator_eval_names.append(clusterer_eval_name)
            else:
                raise ValueError(
                    f"Duplicate evaluation name {clusterer_eval_name} found."
                )

            for n, dataset_name in enumerate(dataset_names[i]):
                for resample in resamples:
                    for split in splits:
                        try:
                            result = ClustererResults().load_from_file(
                                f"{path}/{clusterer_name}/Predictions/"
                                f"{dataset_name}/{split}Resample{resample}.csv",
                                verify_values=verify_results,
                            )
                            clusterer_results.append(result)
                            names.append(clusterer_eval_name)
                            found_estimator = True
                            found_datasets[n] = True

                            if verbose:
                                print(  # noqa: T201
                                    f"Loaded {clusterer_eval_name} on {dataset_name} "
                                    f"{split} resample {resample}."
                                )

                            if result.dataset_name != dataset_name:
                                if verbose:
                                    print(  # noqa: T201
                                        f"File for {clusterer_eval_name} on "
                                        f"{dataset_name} {split} resample {resample} "
                                        f"has a different dataset name than used to "
                                        f"load the file. Setting dataset name to "
                                        f"{dataset_name} from {result.dataset_name}."
                                    )
                                result.dataset_name = dataset_name
                        except FileNotFoundError:
                            msg = (
                                f"Results for {clusterer_eval_name} on {dataset_name} "
                                f"{split} resample {resample} not found."
                            )
                            if error_on_missing:
                                raise FileNotFoundError(msg)
                            elif verbose:
                                print(msg)  # noqa: T201

            if not found_estimator:
                print(f"Clusterer {clusterer_eval_name} not found.")  # noqa: T201

        missing_datasets = [
            dataset
            for dataset, found in zip(dataset_names[i], found_datasets)
            if not found
        ]
        if missing_datasets:
            msg = f"Files for datasets {missing_datasets} not found."
            if error_on_missing:
                raise FileNotFoundError(msg)
            else:
                print("\n\n" + msg)  # noqa: T201

    evaluate_clusterers(
        clusterer_results,
        save_path,
        error_on_missing=error_on_missing,
        eval_name=eval_name,
        estimator_names=names,
    )


def evaluate_regressors(
    regressor_results,
    save_path,
    error_on_missing=True,
    eval_name=None,
    estimator_names=None,
):
    """
    Evaluate multiple regressors on multiple datasets.

    Writes multiple csv files and figures to save_path, one for each statistic
    evaluated. Provides a summary csv file with the average statistic and
    average rank for each regressor.

    Parameters
    ----------
    regressor_results : list of RegressorResults
        The results to evaluate.
    save_path : str
        The path to save the evaluation results to.
    error_on_missing : bool, default=True
        Whether to raise an error if results are missing.
    eval_name : str, default=None
        The name of the evaluation, used in save_path.
    estimator_names : list of str, default=None
        The names of the estimator for each regressor result. If None, uses
        the estimator_name attribute of each regressor result.
    """
    _evaluate_estimators(
        regressor_results,
        RegressorResults.statistics,
        save_path,
        error_on_missing,
        eval_name,
        estimator_names,
    )


def evaluate_regressors_from_file(
    load_paths,
    save_path,
    error_on_missing=True,
    eval_name=None,
    verify_results=True,
    estimator_names=None,
):
    """
    Evaluate multiple regressors on multiple datasets from file.

    Writes multiple csv files and figures to save_path, one for each statistic
    evaluated. Provides a summary csv file with the average statistic and
    average rank for each regressor.

    Parameters
    ----------
    load_paths : list of str
        The paths to the regressor result files to evaluate.
    save_path : str
        The path to save the evaluation results to.
    error_on_missing : bool, default=True
        Whether to raise an error if results are missing.
    eval_name : str, default=None
        The name of the evaluation, used in save_path.
    verify_results : bool, default=True
        If the verification should be performed on the loaded results values.
    estimator_names : list of str, default=None
        The names of the estimator for each regressor result. If None, uses
        the estimator_name attribute of each regressor result.
    """
    regressor_results = []
    for load_path in load_paths:
        try:
            regressor_results.append(
                RegressorResults().load_from_file(
                    load_path, verify_values=verify_results
                )
            )
        except FileNotFoundError:
            if error_on_missing:
                raise FileNotFoundError(f"Results for {load_path} not found.")

    evaluate_regressors(
        regressor_results,
        save_path,
        error_on_missing=error_on_missing,
        eval_name=eval_name,
        estimator_names=estimator_names,
    )


def evaluate_regressors_by_problem(
    load_path,
    regressor_names,
    dataset_names,
    save_path,
    resamples=None,
    load_train_results=False,
    error_on_missing=True,
    eval_name=None,
    verify_results=True,
    verbose=False,
):
    """
    Evaluate multiple regressors on multiple datasets from file using standard paths.

    Finds files using regressor, dataset and resample names. It is expected the
    common tsml-eval file structure of
    {regressor}/Predictions/{dataset}/{split}Resample{resample}.csv is followed.

    Writes multiple csv files and figures to save_path, one for each statistic
    evaluated. Provides a summary csv file with the average statistic and
    average rank for each regressor.

    Parameters
    ----------
    load_path : str or list of str
        The path to the collection of regressor result files to evaluate.
        If load_path is a list, it will load results from each path in the list. It
        is expected that regressor_names and dataset_names are lists of lists with
        the same length as load_path.
    regressor_names : list of str, list of tuple or list of list
        The names of the regressors to evaluate.
        A length 2 tuple containing strings can be used to specify a regressor name to
        load from in the first item and a regressor name to use in the evaluation
        results in the second.
        If load_path is a list, regressor_names must be a list of lists with the same
        length as load_path.
    dataset_names : str, list of str or list of list
        The names of the datasets to evaluate. If a list of strings, each item is the
        name of a dataset. If a string, it is the path to a file containing the names
        of the datasets, one per line.
        If load_path is a list, dataset_names must be a list of lists with the same
        length as load_path.
    save_path : str
        The path to save the evaluation results to.
    resamples : int or list of int, default=None
        The resamples to evaluate. If int, evaluates resamples 0 to resamples-1.
    load_train_results : bool, default=False
        Whether to load train results as well as test results.
    error_on_missing : bool, default=True
        Whether to raise an error if results are missing.
    eval_name : str, default=None
        The name of the evaluation, used in save_path.
    verify_results : bool, default=True
        If the verification should be performed on the loaded results values.
    verbose : bool, default=False
        If verbose output should be printed.
    """
    load_path, regressor_names, dataset_names, resamples = _evaluate_by_problem_init(
        "regressor",
        load_path,
        regressor_names,
        dataset_names,
        resamples,
    )

    if load_train_results:
        splits = ["test", "train"]
    else:
        splits = ["test"]

    regressor_results = []
    estimator_eval_names = []
    names = []
    for i, path in enumerate(load_path):
        found_datasets = np.zeros(len(dataset_names[i]), dtype=bool)

        for regressor_name in regressor_names[i]:
            found_estimator = False

            if isinstance(regressor_name, tuple):
                regressor_eval_name = regressor_name[1]
                regressor_name = regressor_name[0]
            else:
                regressor_eval_name = regressor_name

            if regressor_eval_name not in estimator_eval_names:
                estimator_eval_names.append(regressor_eval_name)
            else:
                raise ValueError(
                    f"Duplicate evaluation name {regressor_eval_name} found."
                )

            for n, dataset_name in enumerate(dataset_names[i]):
                for resample in resamples:
                    for split in splits:
                        try:
                            result = RegressorResults().load_from_file(
                                f"{path}/{regressor_name}/Predictions/"
                                f"{dataset_name}/{split}Resample{resample}.csv",
                                verify_values=verify_results,
                            )
                            regressor_results.append(result)
                            names.append(regressor_eval_name)
                            found_estimator = True
                            found_datasets[n] = True

                            if verbose:
                                print(  # noqa: T201
                                    f"Loaded {regressor_eval_name} on {dataset_name} "
                                    f"{split} resample {resample}."
                                )

                            if result.dataset_name != dataset_name:
                                if verbose:
                                    print(  # noqa: T201
                                        f"File for {regressor_eval_name} on "
                                        f"{dataset_name} {split} resample {resample} "
                                        f"has a different dataset name than used to "
                                        f"load the file. Setting dataset name to "
                                        f"{dataset_name} from {result.dataset_name}."
                                    )
                                result.dataset_name = dataset_name
                        except FileNotFoundError:
                            msg = (
                                f"Results for {regressor_eval_name} on {dataset_name} "
                                f"{split} resample {resample} not found."
                            )
                            if error_on_missing:
                                raise FileNotFoundError(msg)
                            elif verbose:
                                print(msg)  # noqa: T201

            if not found_estimator:
                print(f"Regressor {regressor_eval_name} not found.")  # noqa: T201

        missing_datasets = [
            dataset
            for dataset, found in zip(dataset_names[i], found_datasets)
            if not found
        ]
        if missing_datasets:
            msg = f"Files for datasets {missing_datasets} not found."
            if error_on_missing:
                raise FileNotFoundError(msg)
            else:
                print("\n\n" + msg)  # noqa: T201

    evaluate_regressors(
        regressor_results,
        save_path,
        error_on_missing=error_on_missing,
        eval_name=eval_name,
        estimator_names=names,
    )


def evaluate_forecasters(
    forecaster_results,
    save_path,
    error_on_missing=True,
    eval_name=None,
    estimator_names=None,
):
    """
    Evaluate multiple forecasters on multiple datasets.

    Writes multiple csv files and figures to save_path, one for each statistic
    evaluated. Provides a summary csv file with the average statistic and
    average rank for each forecaster.

    Parameters
    ----------
    forecaster_results : list of ForecasterResults
        The results to evaluate.
    save_path : str
        The path to save the evaluation results to.
    error_on_missing : bool, default=True
        Whether to raise an error if results are missing.
    eval_name : str, default=None
        The name of the evaluation, used in save_path.
    estimator_names : list of str, default=None
        The names of the estimator for each forecaster result. If None, uses
        the estimator_name attribute of each forecaster result.
    """
    _evaluate_estimators(
        forecaster_results,
        ForecasterResults.statistics,
        save_path,
        error_on_missing,
        eval_name,
        estimator_names,
    )


def evaluate_forecasters_from_file(
    load_paths,
    save_path,
    error_on_missing=True,
    eval_name=None,
    verify_results=True,
    estimator_names=None,
):
    """
    Evaluate multiple forecasters on multiple datasets from file.

    Writes multiple csv files and figures to save_path, one for each statistic
    evaluated. Provides a summary csv file with the average statistic and
    average rank for each forecaster.

    Parameters
    ----------
    load_paths : list of str
        The paths to the forecaster result files to evaluate.
    save_path : str
        The path to save the evaluation results to.
    error_on_missing : bool, default=True
        Whether to raise an error if results are missing.
    eval_name : str, default=None
        The name of the evaluation, used in save_path.
    verify_results : bool, default=True
        If the verification should be performed on the loaded results values.
    estimator_names : list of str, default=None
        The names of the estimator for each forecaster result. If None, uses
        the estimator_name attribute of each forecaster result.
    """
    forecaster_results = []
    for load_path in load_paths:
        try:
            forecaster_results.append(
                ForecasterResults().load_from_file(
                    load_path, verify_values=verify_results
                )
            )
        except FileNotFoundError:
            if error_on_missing:
                raise FileNotFoundError(f"Results for {load_path} not found.")

    evaluate_forecasters(
        forecaster_results,
        save_path,
        error_on_missing=error_on_missing,
        eval_name=eval_name,
        estimator_names=estimator_names,
    )


def evaluate_forecasters_by_problem(
    load_path,
    forecaster_names,
    dataset_names,
    save_path,
    resamples=None,
    error_on_missing=True,
    eval_name=None,
    verify_results=True,
    verbose=False,
):
    """
    Evaluate multiple forecasters on multiple datasets from file using standard paths.

    Finds files using forecaster, dataset and resample names. It is expected the
    common tsml-eval file structure of
    {forecaster}/Predictions/{dataset}/{split}Resample{resample}.csv is followed.

    Writes multiple csv files and figures to save_path, one for each statistic
    evaluated. Provides a summary csv file with the average statistic and
    average rank for each forecaster.

    Parameters
    ----------
    load_path : str or list of str
        The path to the collection of forecaster result files to evaluate.
        If load_path is a list, it will load results from each path in the list. It
        is expected that forecaster_names and dataset_names are lists of lists with
        the same length as load_path.
    forecaster_names : list of str, list of tuple or list of list
        The names of the forecasters to evaluate.
        A length 2 tuple containing strings can be used to specify a forecaster name to
        load from in the first item and a forecaster name to use in the evaluation
        results in the second.
        If load_path is a list, regressor_names must be a list of lists with the same
        length as load_path.
    dataset_names : str, list of str or list of list
        The names of the datasets to evaluate. If a list of strings, each item is the
        name of a dataset. If a string, it is the path to a file containing the names
        of the datasets, one per line.
        If load_path is a list, dataset_names must be a list of lists with the same
        length as load_path.
    save_path : str
        The path to save the evaluation results to.
    resamples : int or list of int, default=None
        The resamples to evaluate. If int, evaluates resamples 0 to resamples-1.
    error_on_missing : bool, default=True
        Whether to raise an error if results are missing.
    eval_name : str, default=None
        The name of the evaluation, used in save_path.
    verify_results : bool, default=True
        If the verification should be performed on the loaded results values.
    verbose : bool, default=False
        If verbose output should be printed.
    """
    load_path, forecaster_names, dataset_names, resamples = _evaluate_by_problem_init(
        "forecaster",
        load_path,
        forecaster_names,
        dataset_names,
        resamples,
    )

    forecaster_results = []
    estimator_eval_names = []
    names = []
    for i, path in enumerate(load_path):
        found_datasets = np.zeros(len(dataset_names[i]), dtype=bool)

        for forecaster_name in forecaster_names[i]:
            found_estimator = False

            if isinstance(forecaster_name, tuple):
                forecaster_eval_name = forecaster_name[1]
                forecaster_name = forecaster_name[0]
            else:
                forecaster_eval_name = forecaster_name

            if forecaster_eval_name not in estimator_eval_names:
                estimator_eval_names.append(forecaster_eval_name)
            else:
                raise ValueError(
                    f"Duplicate evaluation name {forecaster_eval_name} found."
                )

            for n, dataset_name in enumerate(dataset_names[i]):
                for resample in resamples:
                    try:
                        result = ForecasterResults().load_from_file(
                            f"{path}/{forecaster_name}/Predictions/"
                            f"{dataset_name}/testResample{resample}.csv",
                            verify_values=verify_results,
                        )
                        forecaster_results.append(result)
                        names.append(forecaster_eval_name)
                        found_estimator = True
                        found_datasets[n] = True

                        if verbose:
                            print(  # noqa: T201
                                f"Loaded {forecaster_eval_name} on {dataset_name} "
                                f"resample {resample}."
                            )

                        if result.dataset_name != dataset_name:
                            if verbose:
                                print(  # noqa: T201
                                    f"File for {forecaster_eval_name} on "
                                    f"{dataset_name} resample {resample} "
                                    f"has a different dataset name than used to "
                                    f"load the file. Setting dataset name to "
                                    f"{dataset_name} from {result.dataset_name}."
                                )
                            result.dataset_name = dataset_name
                    except FileNotFoundError:
                        msg = (
                            f"Results for {forecaster_eval_name} on {dataset_name} "
                            f"resample {resample} not found."
                        )
                        if error_on_missing:
                            raise FileNotFoundError(msg)
                        elif verbose:
                            print(msg)  # noqa: T201

            if not found_estimator:
                print(f"Forecaster {forecaster_eval_name} not found.")  # noqa: T201

        missing_datasets = [
            dataset
            for dataset, found in zip(dataset_names[i], found_datasets)
            if not found
        ]
        if missing_datasets:
            msg = f"Files for datasets {missing_datasets} not found."
            if error_on_missing:
                raise FileNotFoundError(msg)
            else:
                print("\n\n" + msg)  # noqa: T201

    evaluate_forecasters(
        forecaster_results,
        save_path,
        error_on_missing=error_on_missing,
        eval_name=eval_name,
        estimator_names=names,
    )


def _evaluate_by_problem_init(
    type, load_path, estimator_names, dataset_names, resamples
):
    if isinstance(load_path, str):
        load_path = [load_path]
    elif not isinstance(load_path, list):
        raise TypeError("load_path must be a str or list of str.")

    if isinstance(estimator_names[0], (str, tuple)):
        estimator_names = [estimator_names]
    elif not isinstance(estimator_names[0], list):
        raise TypeError(f"{type}_names must be a str, tuple or list of str or tuple.")

    if isinstance(dataset_names, str):
        with open(dataset_names) as f:
            dataset_names = f.readlines()
            dataset_names = [[d.strip() for d in dataset_names]] * len(load_path)
    elif isinstance(dataset_names[0], str):
        dataset_names = [dataset_names] * len(load_path)
    elif not isinstance(dataset_names[0], list):
        raise TypeError("dataset_names must be a str or list of str.")

    if len(load_path) != len(estimator_names) or len(load_path) != len(dataset_names):
        raise ValueError(
            f"load_path, {type}_names and dataset_names must be the same length."
        )

    if resamples is None:
        resamples = [""]
    elif isinstance(resamples, int):
        resamples = [str(i) for i in range(resamples)]
    else:
        resamples = [str(resample) for resample in resamples]

    return load_path, estimator_names, dataset_names, resamples


def _evaluate_estimators(
    estimator_results,
    statistics,
    save_path,
    error_on_missing,
    eval_name,
    estimator_names,
):
    save_path = save_path + "/" + eval_name + "/"

    estimators = set()
    datasets = set()
    resamples = set()
    has_test = False
    has_train = False

    results_dict = _create_results_dictionary(estimator_results, estimator_names)

    if eval_name is None:
        dt = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        eval_name = f"{estimator_results[0].__class__.__name__}Evaluation {dt}"

    for estimator_name in results_dict:
        estimators.add(estimator_name)
        for dataset_name in results_dict[estimator_name]:
            datasets.add(dataset_name)
            for split in results_dict[estimator_name][dataset_name]:
                split_fail = False
                if split == "train":
                    has_train = True
                elif split == "test":
                    has_test = True
                else:
                    split_fail = True

                for resample in results_dict[estimator_name][dataset_name][split]:
                    if split_fail:
                        raise ValueError(
                            "Results must have a split of either 'train' or 'test' "
                            f"to be evaluated. Missing for {estimator_name} on "
                            f"{dataset_name} resample {resample}."
                        )

                    if resample is not None:
                        resamples.add(resample)
                    else:
                        raise ValueError(
                            "Results must have a resample_id to be evaluated. "
                            f"Missing for {estimator_name} on {dataset_name} "
                            f"{split} resample {resample}."
                        )

    estimators = sorted(list(estimators))
    datasets = sorted(list(datasets))
    resamples = sorted(list(resamples))
    has_dataset_train = np.zeros(
        (len(estimators), len(datasets), len(resamples)), dtype=bool
    )
    has_dataset_test = np.zeros(
        (len(estimators), len(datasets), len(resamples)), dtype=bool
    )

    for estimator_name in results_dict:
        for dataset_name in results_dict[estimator_name]:
            for split in results_dict[estimator_name][dataset_name]:
                for resample in results_dict[estimator_name][dataset_name][split]:
                    if split == "train":
                        has_dataset_train[estimators.index(estimator_name)][
                            datasets.index(dataset_name)
                        ][resamples.index(resample)] = True
                    elif split == "test":
                        has_dataset_test[estimators.index(estimator_name)][
                            datasets.index(dataset_name)
                        ][resamples.index(resample)] = True

    msg = "\n\n"
    missing = False
    splits = []

    if has_train:
        splits.append("train")
        for (i, n, j), present in np.ndenumerate(has_dataset_train):
            if not present:
                msg += (
                    f"Estimator {estimators[i]} is missing train results for "
                    f"{datasets[n]} resample {resamples[j]}.\n"
                )
                missing = True

    if has_test:
        splits.append("test")
        for (i, n, j), present in np.ndenumerate(has_dataset_test):
            if not present:
                msg += (
                    f"Estimator {estimators[i]} is missing test results for "
                    f"{datasets[n]} resample {resamples[j]}.\n"
                )
                missing = True

    if missing:
        if error_on_missing:
            print(msg + "\n")  # noqa: T201
            raise ValueError("Missing results, exiting evaluation.")
        else:
            if has_test and has_train:
                has_both = has_dataset_train.all(axis=(0, 2)) & has_dataset_test.all(
                    axis=(0, 2)
                )
                datasets = [dataset for dataset, has in zip(datasets, has_both) if has]
            elif has_test:
                datasets = [
                    dataset
                    for dataset, has in zip(datasets, has_dataset_test.all(axis=(0, 2)))
                    if has
                ]
            else:
                datasets = [
                    dataset
                    for dataset, has in zip(
                        datasets, has_dataset_train.all(axis=(0, 2))
                    )
                    if has
                ]

            msg += "\nMissing results, continuing evaluation with available datasets.\n"
            print(msg)  # noqa: T201
    else:
        msg += "All results present, continuing evaluation.\n"
        print(msg)  # noqa: T201

    print(f"Estimators ({len(estimators)}): {estimators}\n")  # noqa: T201
    print(f"Datasets ({len(datasets)}): {datasets}\n")  # noqa: T201
    print(f"Resamples ({len(resamples)}): {resamples}\n")  # noqa: T201

    stats = []
    for var, (stat, ascending, time) in statistics.items():
        for split in splits:
            average, rank = _create_directory_for_statistic(
                estimators,
                datasets,
                resamples,
                split,
                results_dict,
                stat,
                ascending,
                time,
                var,
                save_path,
                eval_name,
            )
            stats.append((average, rank, stat, ascending, split))

    _summary_evaluation(stats, estimators, save_path, eval_name)


def _create_results_dictionary(estimator_results, estimator_names):
    results_dict = {}

    for i, estimator_result in enumerate(estimator_results):
        name = (
            estimator_result.estimator_name
            if estimator_names is None
            else estimator_names[i]
        )

        if results_dict.get(name) is None:
            results_dict[name] = {}

        if results_dict[name].get(estimator_result.dataset_name) is None:
            results_dict[name][estimator_result.dataset_name] = {}

        if (
            results_dict[name][estimator_result.dataset_name].get(
                estimator_result.split.lower()
            )
            is None
        ):
            results_dict[name][estimator_result.dataset_name][
                estimator_result.split.lower()
            ] = {}

        results_dict[name][estimator_result.dataset_name][
            estimator_result.split.lower()
        ][estimator_result.resample_id] = estimator_result

    return results_dict


def _create_directory_for_statistic(
    estimators,
    datasets,
    resamples,
    split,
    results_dict,
    statistic_name,
    higher_better,
    is_timing,
    variable_name,
    save_path,
    eval_name,
):
    os.makedirs(f"{save_path}/{statistic_name}/all_resamples/", exist_ok=True)

    average_stats = np.zeros((len(datasets), len(estimators)))

    for i, estimator_name in enumerate(estimators):
        est_stats = np.zeros((len(datasets), len(resamples)))

        for n, dataset_name in enumerate(datasets):
            for j, resample in enumerate(resamples):
                er = results_dict[estimator_name][dataset_name][split][resample]
                er.calculate_statistics()
                est_stats[n, j] = (
                    er.__dict__[variable_name]
                    if not is_timing
                    else (
                        time_to_milliseconds(er.__dict__[variable_name], er.time_unit)
                    )
                )

            average_stats[n, i] = np.mean(est_stats[n, :])

        with open(
            f"{save_path}/{statistic_name}/all_resamples/"
            f"{estimator_name}_{statistic_name.lower()}.csv",
            "w",
        ) as file:
            file.write(f"Resamples:,{','.join([str(j) for j in resamples])}\n")
            for n, dataset_name in enumerate(datasets):
                file.write(
                    f"{dataset_name},{','.join([str(j) for j in est_stats[n]])}\n"
                )

    with open(
        f"{save_path}/{statistic_name}/{statistic_name.lower()}_mean.csv", "w"
    ) as file:
        file.write(f"Estimators:,{','.join(estimators)}\n")
        for i, dataset_name in enumerate(datasets):
            file.write(
                f"{dataset_name},{','.join([str(n) for n in average_stats[i]])}\n"
            )

    ranks = np.apply_along_axis(
        lambda x: rank_array(x, higher_better=higher_better), 1, average_stats
    )

    with open(
        f"{save_path}/{statistic_name}/{statistic_name.lower()}_ranks.csv", "w"
    ) as file:
        file.write(f"Estimators:,{','.join(estimators)}\n")
        for i, dataset_name in enumerate(datasets):
            file.write(f"{dataset_name},{','.join([str(n) for n in ranks[i]])}\n")

    p_values = wilcoxon_test(average_stats, estimators, lower_better=not higher_better)
    with open(
        f"{save_path}/{statistic_name}/{statistic_name.lower()}_p_values.csv", "w"
    ) as file:
        file.write(f"Estimators:,{','.join(estimators)}\n")
        for i, estimator_name in enumerate(estimators):
            file.write(f"{estimator_name},{','.join([str(n) for n in p_values[i]])}\n")

    try:
        _figures_for_statistic(
            average_stats,
            estimators,
            statistic_name,
            higher_better,
            save_path,
            eval_name,
        )
    except ValueError as e:
        warnings.warn(
            f"Error during figure creation for {statistic_name}: {e}",
            stacklevel=2,
        )

    return average_stats, ranks


def _figures_for_statistic(
    scores, estimators, statistic_name, higher_better, save_path, eval_name
):
    os.makedirs(f"{save_path}/{statistic_name}/figures/", exist_ok=True)

    cd, _ = plot_critical_difference(scores, estimators, lower_better=not higher_better)
    cd.savefig(
        f"{save_path}/{statistic_name}/figures/"
        f"{eval_name}_{statistic_name.lower()}_critical_difference.pdf",
        bbox_inches="tight",
    )
    pickle.dump(
        cd,
        open(
            f"{save_path}/{statistic_name}/figures/"
            f"{eval_name}_{statistic_name.lower()}_critical_difference.pickle",
            "wb",
        ),
    )
    plt.close()

    box, _ = plot_boxplot_median(scores, estimators)
    box.savefig(
        f"{save_path}/{statistic_name}/figures/"
        f"{eval_name}_{statistic_name.lower()}_boxplot.pdf",
        bbox_inches="tight",
    )
    pickle.dump(
        box,
        open(
            f"{save_path}/{statistic_name}/figures/"
            f"{eval_name}_{statistic_name.lower()}_boxplot.pickle",
            "wb",
        ),
    )
    plt.close()

    for i, est1 in enumerate(estimators):
        for n, est2 in enumerate(estimators):
            if i == n:
                continue

            os.makedirs(
                f"{save_path}/{statistic_name}/figures/scatters/{est1}/", exist_ok=True
            )

            scatter, _ = plot_pairwise_scatter(scores[:, i], scores[:, n], est1, est2)
            scatter.savefig(
                f"{save_path}/{statistic_name}/figures/scatters/{est1}/"
                f"{eval_name}_{statistic_name.lower()}_scatter_{est1}_{est2}.pdf",
                bbox_inches="tight",
            )
            pickle.dump(
                scatter,
                open(
                    f"{save_path}/{statistic_name}/figures/scatters/{est1}/"
                    f"{eval_name}_{statistic_name.lower()}_scatter_{est1}_{est2}"
                    f".pickle",
                    "wb",
                ),
            )
            plt.close()


def _summary_evaluation(stats, estimators, save_path, eval_name):
    with open(f"{save_path}/{eval_name}_summary.csv", "w") as file:
        for stat in stats:
            avg_stat = np.mean(stat[0], axis=0)
            avg_rank = np.mean(stat[1], axis=0)
            sorted_indices = [
                i
                for i in sorted(
                    range(len(avg_rank)),
                    key=lambda x: (
                        avg_rank[x],
                        -avg_stat[x] if stat[3] else avg_stat[x],
                    ),
                )
            ]

            file.write(
                f"{stat[4]}{stat[2]},"
                f"{','.join([estimators[i] for i in sorted_indices])}\n"
            )
            file.write(
                f"{stat[4]}{stat[2]}Mean,"
                f"{','.join([str(n) for n in avg_stat[sorted_indices]])}\n"
            )
            file.write(
                f"{stat[4]}{stat[2]}AvgRank,"
                f"{','.join([str(n) for n in avg_rank[sorted_indices]])}\n\n"
            )



"""Test dataset utilities."""

import os

import pytest
from aeon.datasets import load_from_tsfile

from tsml_eval.datasets._test_data._data_sizes import DATA_TEST_SIZES, DATA_TRAIN_SIZES
from tsml_eval.testing.testing_utils import _TEST_DATA_PATH, _TEST_OUTPUT_PATH
from tsml_eval.utils.datasets import copy_dataset_ts_files, save_merged_dataset_splits


def test_copy_dataset_ts_files():
    """Test copying .ts dataset files."""
    copy_path = f"{_TEST_OUTPUT_PATH}/datasets/"

    copy_dataset_ts_files(
        f"{_TEST_DATA_PATH}/_test_data/test_datalist.txt",
        _TEST_DATA_PATH,
        copy_path,
    )

    assert os.path.exists(f"{copy_path}/MinimalChinatown/MinimalChinatown_TRAIN.ts")
    assert os.path.exists(f"{copy_path}/MinimalChinatown/MinimalChinatown_TEST.ts")
    assert os.path.exists(f"{copy_path}/MinimalGasPrices/MinimalGasPrices_TRAIN.ts")
    assert os.path.exists(f"{copy_path}/MinimalGasPrices/MinimalGasPrices_TEST.ts")
    assert os.path.exists(
        f"{_TEST_DATA_PATH}/MinimalChinatown/MinimalChinatown_TRAIN.ts"
    )
    assert os.path.exists(
        f"{_TEST_DATA_PATH}/MinimalChinatown/MinimalChinatown_TEST.ts"
    )
    assert os.path.exists(
        f"{_TEST_DATA_PATH}/MinimalGasPrices/MinimalGasPrices_TRAIN.ts"
    )
    assert os.path.exists(
        f"{_TEST_DATA_PATH}/MinimalGasPrices/MinimalGasPrices_TEST.ts"
    )

    os.remove(f"{copy_path}/MinimalChinatown/MinimalChinatown_TRAIN.ts")
    os.remove(f"{copy_path}/MinimalChinatown/MinimalChinatown_TEST.ts")
    os.remove(f"{copy_path}/MinimalGasPrices/MinimalGasPrices_TRAIN.ts")
    os.remove(f"{copy_path}/MinimalGasPrices/MinimalGasPrices_TEST.ts")


@pytest.mark.parametrize(
    "save_path", [None, f"{_TEST_OUTPUT_PATH}/datasets/merged2/destination/"]
)
def test_save_merged_dataset_splits(save_path):
    """Test merging and saving .ts dataset files."""
    copy_path = (
        f"{_TEST_OUTPUT_PATH}/datasets/merged/"
        if save_path is None
        else f"{_TEST_OUTPUT_PATH}/datasets/merged2/"
    )

    copy_dataset_ts_files(
        ["MinimalChinatown"],
        _TEST_DATA_PATH,
        copy_path,
    )

    save_merged_dataset_splits(
        copy_path,
        "MinimalChinatown",
        save_path,
    )

    if save_path is None:
        save_path = copy_path

    assert os.path.exists(f"{copy_path}/MinimalChinatown/MinimalChinatown_TRAIN.ts")
    assert os.path.exists(f"{copy_path}/MinimalChinatown/MinimalChinatown_TEST.ts")
    assert os.path.exists(f"{save_path}/MinimalChinatown/MinimalChinatown.ts")

    X, y = load_from_tsfile(f"{save_path}/MinimalChinatown/MinimalChinatown.ts")

    new_len = DATA_TRAIN_SIZES["MinimalChinatown"] + DATA_TEST_SIZES["MinimalChinatown"]
    assert X.shape[0] == new_len
    assert y.shape[0] == new_len

    os.remove(f"{copy_path}/MinimalChinatown/MinimalChinatown_TRAIN.ts")
    os.remove(f"{copy_path}/MinimalChinatown/MinimalChinatown_TEST.ts")
    os.remove(f"{save_path}/MinimalChinatown/MinimalChinatown.ts")



import numpy as np
from sklearn.utils.validation import check_random_state
from aeon.classification.base import BaseClassifier

from tsml_eval._wip.tschief._splitters import generate_boss_transforms
from tsml_eval._wip.tschief.tschiefnode import TsChiefNode

__author__ = ["GuiArcencio"]
__all__ = ["TsChief"]


class TsChief(BaseClassifier):
    """Unfinished implementation of the TS-CHIEF classifier."""

    _tags = {
        "X_inner_mtype": "numpy3D",
        "capability:multivariate": True,
    }

    def __init__(
        self,
        n_trees=500,
        n_dictionary=100,
        n_interval=100,
        n_distance=5,
        n_boss_transformations=1000,
        random_state=None,
    ):
        self.n_trees = n_trees
        self.n_dictionary = n_dictionary
        self.n_interval = n_interval
        self.n_distance = n_distance
        self.n_boss_transformations = n_boss_transformations
        self.random_state = random_state

        super().__init__()

    def _fit(self, X, y):
        self._class_dtype = y.dtype
        self.random_state = check_random_state(self.random_state)

        sfas, X_boss = generate_boss_transforms(
            X, self.n_boss_transformations, self.random_state
        )
        self.trees_ = []
        for _ in range(self.n_trees):
            self.trees_.append(
                TsChiefNode(
                    n_dictionary=self.n_dictionary,
                    n_distance=self.n_distance,
                    n_interval=self.n_interval,
                    random_state=self.random_state,
                ).fit(X, y, X_boss, sfas)
            )

        return self

    def _predict(self, X):
        samples = X.shape[0]

        votes = np.empty((samples, self.n_trees), dtype=self._class_dtype)
        preds = np.empty(samples, dtype=self._class_dtype)

        for i, tree in enumerate(self.trees_):
            votes[:, i] = tree.predict(X)

        for i in range(samples):
            classes, counts = np.unique(votes[i], return_counts=True)
            max_idx = np.argmax(counts)
            preds[i] = classes[max_idx]

        return preds



"""Files for distance-based clustering publication."""

__all__ = [
    "_set_distance_clusterer",
    "_run_experiment",
    "distance_based_clusterers",
]

from tsml_eval.publications.y2023.distance_based_clustering.run_distance_experiments import (  # noqa: E501
    _run_experiment,
)
from tsml_eval.publications.y2023.distance_based_clustering.set_distance_clusterer import (  # noqa: E501
    _set_distance_clusterer,
    distance_based_clusterers,
)



"""Class for storing and loading results from a classification experiment."""

import numpy as np
from numpy.testing import assert_allclose
from sklearn.metrics import (
    accuracy_score,
    balanced_accuracy_score,
    f1_score,
    log_loss,
    roc_auc_score,
)

from tsml_eval.evaluation.storage.estimator_results import EstimatorResults
from tsml_eval.utils.results_writing import write_classification_results


class ClassifierResults(EstimatorResults):
    """
    A class for storing and managing results from classification experiments.

    This class provides functionalities for storing classification results,
    including predictions, probabilities, and various performance metrics.
    It extends the `EstimatorResults` class, inheriting its base functionalities.

    Parameters
    ----------
    dataset_name : str, default="N/A"
        Name of the dataset used.
    classifier_name : str, default="N/A"
        Name of the classifier used.
    split : str, default="N/A"
        Type of data split used, i.e. "train" or "test".
    resample_id : int or None, default=None
        Random seed used for the data resample, with 0 usually being the original data.
    time_unit : str, default="nanoseconds"
        Time measurement used for other fields.
    description : str, default=""
        Additional description of the classification experiment. Appended to the end
        of the first line of the results file.
    parameters : str, default="No parameter info"
        Information about parameters used in the classifier and other build information.
        Written to the second line of the results file.
    fit_time : float, default=-1.0
        Time taken fitting the model.
    predict_time : float, default=-1.0
        Time taken making predictions.
    benchmark_time : float, default=-1.0
        Time taken to run a simple benchmark function. In tsml-eval experiments, this
        is the time spent to sort 1,000 (seeded) random numpy arrays of size 20,000.
    memory_usage : float, default=-1.0
        Memory usage during the experiment. In tsml-eval experiments, this is the peak
        memory usage during the fit method.
    n_classes : int or None, default=None
        Number of classes in the dataset.
    error_estimate_method : str, default="N/A"
        Method used for train error/accuracy estimation (i.e. 10-fold CV, OOB error).
    error_estimate_time : float, default=-1.0
        Time taken for train error/accuracy estimation.
    build_plus_estimate_time : float, default=-1.0
        Total time for building the classifier and estimating error/accuracy on the
        train set. For certain methods this can be different from the sum of fit_time
        and error_estimate_time.
    class_labels : array-like or None, default=None
        Actual class labels.
    predictions : array-like or None, default=None
        Predicted class labels.
    probabilities : array-like or None, default=None
        Predicted class probabilities.
    pred_times : array-like or None, default=None
        Prediction times for each case.
    pred_descriptions : list of str or None, default=None
        Descriptions for each prediction.

    Attributes
    ----------
    n_cases : int or None
        Number of cases in the dataset.
    accuracy : float or None
        Accuracy of the classifier.
    balanced_accuracy : float or None
        Balanced accuracy of the classifier.
    f1_score : float or None
        F1 score of the classifier.
    log_loss : float or None
        Negative log likelihood of the classifier.
    auroc_score : float or None
        Mean area under the ROC curve of the classifier.

    Examples
    --------
    >>> from tsml_eval.evaluation.storage import ClassifierResults
    >>> from tsml_eval.testing.testing_utils import _TEST_RESULTS_PATH
    >>> cr = ClassifierResults().load_from_file(
    ...     _TEST_RESULTS_PATH +
    ...     "/classification/ROCKET/Predictions/Chinatown/testResample0.csv"
    ... )
    >>> cr.calculate_statistics()
    >>> cr.accuracy
    0.9795918367346939
    """

    def __init__(
        self,
        dataset_name="N/A",
        classifier_name="N/A",
        split="N/A",
        resample_id=None,
        time_unit="nanoseconds",
        description="",
        parameters="No parameter info",
        fit_time=-1.0,
        predict_time=-1.0,
        benchmark_time=-1.0,
        memory_usage=-1.0,
        n_classes=None,
        error_estimate_method="N/A",
        error_estimate_time=-1.0,
        build_plus_estimate_time=-1.0,
        class_labels=None,
        predictions=None,
        probabilities=None,
        pred_times=None,
        pred_descriptions=None,
    ):
        # Line 1
        self.classifier_name = classifier_name

        # Line 3
        self.n_classes = n_classes
        self.train_estimate_method = error_estimate_method
        self.train_estimate_time = error_estimate_time
        self.fit_and_estimate_time = build_plus_estimate_time

        # Results
        self.class_labels = class_labels
        self.predictions = predictions
        self.probabilities = probabilities
        self.pred_times = pred_times
        self.pred_descriptions = pred_descriptions

        self.n_cases = None

        self.accuracy = None
        self.balanced_accuracy = None
        self.auroc_score = None
        self.log_loss = None
        self.f1_score = None

        super().__init__(
            dataset_name=dataset_name,
            estimator_name=classifier_name,
            split=split,
            resample_id=resample_id,
            time_unit=time_unit,
            description=description,
            parameters=parameters,
            fit_time=fit_time,
            predict_time=predict_time,
            benchmark_time=benchmark_time,
            memory_usage=memory_usage,
        )

    # var_name: (display_name, higher is better, is timing)
    statistics = {
        "accuracy": ("Accuracy", True, False),
        "balanced_accuracy": ("BalAcc", True, False),
        "auroc_score": ("AUROC", True, False),
        "log_loss": ("LogLoss", False, False),
        "f1_score": ("F1", True, False),
        **EstimatorResults.statistics,
    }

    def save_to_file(self, file_path, full_path=True):
        """
        Write the classifier results into a file format used by tsml.

        Parameters
        ----------
        file_path : str
            Path to write the results file to or the directory to build the default file
            structure if full_path is False.
        full_path : boolean, default=True
            If True, results are written directly to the directory passed in file_path.
            If False, then a standard file structure using the classifier and dataset
            names is created and used to write the results file.
        """
        self.infer_size()

        if self.accuracy is None:
            self.accuracy = accuracy_score(self.class_labels, self.predictions)

        write_classification_results(
            self.predictions,
            self.probabilities,
            self.class_labels,
            self.estimator_name,
            self.dataset_name,
            file_path,
            full_path=full_path,
            split=self.split,
            resample_id=self.resample_id,
            time_unit=self.time_unit,
            first_line_comment=self.description,
            parameter_info=self.parameter_info,
            accuracy=self.accuracy,
            fit_time=self.fit_time,
            predict_time=self.predict_time,
            benchmark_time=self.benchmark_time,
            memory_usage=self.memory_usage,
            n_classes=self.n_classes,
            train_estimate_method=self.train_estimate_method,
            train_estimate_time=self.train_estimate_time,
            fit_and_estimate_time=self.fit_and_estimate_time,
        )

    def load_from_file(self, file_path, verify_values=True):
        """
        Load classifier results from a specified file.

        This method reads a file containing classifier results and reconstructs the
        ClassifierResults object. It calculates performance statistics and
        verifies values based on the loaded data.

        Parameters
        ----------
        file_path : str
            The path to the file from which classifier results should be loaded. The
            file should be a tsml formatted classifier results file.
        verify_values : bool, default=True
            If the method should perform verification of the loaded values.

        Returns
        -------
        self : ClassifierResults
            The same ClassifierResults object with loaded results.
        """
        cr = load_classifier_results(file_path, verify_values=verify_values)
        self.__dict__.update(cr.__dict__)
        return self

    def calculate_statistics(self, overwrite=False):
        """
        Calculate various performance statistics based on the classifier results.

        This method computes various performance metrics, such as accuracy, F1 score,
        and others, based on the classifiers output.

        Parameters
        ----------
        overwrite : bool, default=False
            If the function should overwrite the current values when they are not None.
        """
        self.infer_size(overwrite=overwrite)

        if self.accuracy is None or overwrite:
            self.accuracy = accuracy_score(self.class_labels, self.predictions)
        if self.balanced_accuracy is None or overwrite:
            self.balanced_accuracy = balanced_accuracy_score(
                self.class_labels, self.predictions
            )
        if self.log_loss is None or overwrite:
            import warnings

            warnings.filterwarnings(
                "ignore",
                message="The y_pred values do not sum to one. Starting from 1.5 "
                "thiswill result in an error.",
            )
            self.log_loss = log_loss(
                self.class_labels,
                self.probabilities,
            )
        if self.auroc_score is None or overwrite:
            self.auroc_score = roc_auc_score(
                self.class_labels,
                self.probabilities[:, 1] if self.n_classes == 2 else self.probabilities,
                average="weighted",
                multi_class="ovr",
            )
        if self.f1_score is None or overwrite:
            self.f1_score = f1_score(
                self.class_labels, self.predictions, average="weighted"
            )

    def infer_size(self, overwrite=False):
        """
        Infer and return the size of the dataset used in the results.

        This method estimates the size of the dataset that was used for the estimator,
        based on the results data.

        Parameters
        ----------
        overwrite : bool, default=False
            If the function should overwrite the current values when they are not None.
        """
        if self.n_cases is None or overwrite:
            self.n_cases = len(self.class_labels)
        if self.n_classes is None or overwrite:
            self.n_classes = len(self.probabilities[0])


def load_classifier_results(file_path, calculate_stats=True, verify_values=True):
    """
    Load and return classifier results from a specified file.

    This function reads a file containing classifier results and reconstructs the
    ClassifierResults object. It optionally calculates performance statistics and
    verifies values based on the loaded data.

    Parameters
    ----------
    file_path : str
        The path to the file from which classifier results should be loaded. The file
        should be a tsml formatted classifier results file.
    calculate_stats : bool, default=True
        Whether to calculate performance statistics from the loaded results.
    verify_values : bool, default=True
        If the function should perform verification of the loaded values.

    Returns
    -------
    cr : ClassifierResults
        A ClassifierResults object containing the results loaded from the file.
    """
    with open(file_path) as file:
        lines = file.readlines()

        line1 = lines[0].split(",")
        line3 = lines[2].split(",")
        acc = float(line3[0])
        n_classes = int(line3[5])
        n_cases = len(lines) - 3

        line_size = len(lines[3].split(","))

        class_labels = np.zeros(n_cases)
        predictions = np.zeros(n_cases)
        probabilities = np.zeros((n_cases, n_classes))

        if line_size > 4 + n_classes:
            pred_times = np.zeros(n_cases)
        else:
            pred_times = None

        if line_size > 6 + n_classes:
            pred_descriptions = []
        else:
            pred_descriptions = None

        for i in range(0, n_cases):
            line = lines[i + 3].split(",")
            class_labels[i] = int(line[0])
            predictions[i] = int(line[1])

            for j in range(0, n_classes):
                probabilities[i, j] = float(line[3 + j])

            if pred_times is not None:
                pred_times[i] = float(line[4 + n_classes])

            if pred_descriptions is not None:
                pred_descriptions.append(",".join(line[6 + n_classes :]).strip())

        # compatability with old results files
        if len(line3) > 6:
            error_estimate_method = line3[6]
            error_estimate_time = float(line3[7])
            build_plus_estimate_time = float(line3[8])
        else:
            error_estimate_method = "N/A"
            error_estimate_time = -1.0
            build_plus_estimate_time = -1.0

    cr = ClassifierResults(
        dataset_name=line1[0],
        classifier_name=line1[1],
        split=line1[2],
        resample_id=None if line1[3] == "None" else int(line1[3]),
        time_unit=line1[4].lower(),
        description=",".join(line1[5:]).strip(),
        parameters=lines[1].strip(),
        fit_time=float(line3[1]),
        predict_time=float(line3[2]),
        benchmark_time=float(line3[3]),
        memory_usage=float(line3[4]),
        n_classes=n_classes,
        error_estimate_method=error_estimate_method,
        error_estimate_time=error_estimate_time,
        build_plus_estimate_time=build_plus_estimate_time,
        class_labels=class_labels,
        predictions=predictions,
        probabilities=probabilities,
        pred_times=pred_times,
        pred_descriptions=pred_descriptions,
    )

    if calculate_stats:
        cr.calculate_statistics()

    if verify_values:
        cr.infer_size(overwrite=True)
        assert cr.n_cases == n_cases
        assert cr.n_classes == n_classes

        assert_allclose(probabilities.sum(axis=1), 1, rtol=1e-5)

        if calculate_stats:
            assert cr.accuracy == acc

    return cr



"""Test testing utilities."""

import pytest

from tsml_eval.testing.testing_utils import (
    _check_set_method,
    _check_set_method_results,
    suppress_output,
)


def test_check_set_method_fail():
    """Test _check_set_method_ failure state."""
    with pytest.raises(ModuleNotFoundError):
        _check_set_method(
            _test_set_method_fail,
            ["a", "b", "c"],
            {},
            [],
        )


def _test_set_method_fail(estimator_alias):
    raise ModuleNotFoundError(
        "Just imagine you imported a module which does not exist when "
        f"importing {estimator_alias}."
    )


def test_check_set_method_results_fail():
    """Test _check_set_method_results failure state."""
    with pytest.raises(ValueError, match=r"missing entries: \['b'\]"):
        _check_set_method_results({"a": True, "b": False, "c": True})


@suppress_output(suppress_stdout=False, suppress_stderr=False)
def test_suppress_output_false():
    """Test suppress_output method with False inputs."""
    pass



"""Tests for the results repair utilities."""

import os

import pytest

from tsml_eval.testing.testing_utils import _TEST_OUTPUT_PATH, _TEST_RESULTS_PATH
from tsml_eval.utils.results_repair import fix_broken_second_line
from tsml_eval.utils.results_validation import validate_results_file


@pytest.mark.parametrize(
    "path",
    [
        [_TEST_RESULTS_PATH + "/regression/regressionResultsFile.csv", 1],
        [_TEST_RESULTS_PATH + "/broken/brokenRegressionResultsFile.csv", 2],
    ],
)
def test_fix_broken_second_line(path):
    """Test that the second line of a broken results file is fixed."""
    fix_broken_second_line(path[0], f"{_TEST_OUTPUT_PATH}/secondLineTest{path[1]}.csv")

    assert validate_results_file(f"{_TEST_OUTPUT_PATH}/secondLineTest{path[1]}.csv")

    # validate again while overwriting the original file
    fix_broken_second_line(f"{_TEST_OUTPUT_PATH}/secondLineTest{path[1]}.csv")

    assert validate_results_file(f"{_TEST_OUTPUT_PATH}/secondLineTest{path[1]}.csv")

    os.remove(f"{_TEST_OUTPUT_PATH}/secondLineTest{path[1]}.csv")


def test_fix_broken_second_line_invalid_third_line():
    """Test that an error is raised if the third line is broken."""
    path = _TEST_RESULTS_PATH + "/broken/brokenResultsFileLine3.csv"

    with pytest.raises(ValueError, match="No valid third line"):
        fix_broken_second_line(path, f"{_TEST_OUTPUT_PATH}/secondLineTest3.csv")



"""Experiment functions."""

__all__ = [
    "run_classification_experiment",
    "load_and_run_classification_experiment",
    "run_clustering_experiment",
    "load_and_run_clustering_experiment",
    "run_forecasting_experiment",
    "load_and_run_forecasting_experiment",
    "run_regression_experiment",
    "load_and_run_regression_experiment",
    "get_classifier_by_name",
    "get_clusterer_by_name",
    "get_forecaster_by_name",
    "get_regressor_by_name",
    "classification_cross_validation",
    "classification_cross_validation_folds",
    "regression_cross_validation",
    "regression_cross_validation_folds",
]

from tsml_eval.experiments.cross_validation import (
    classification_cross_validation,
    classification_cross_validation_folds,
    regression_cross_validation,
    regression_cross_validation_folds,
)
from tsml_eval.experiments.experiments import (
    load_and_run_classification_experiment,
    load_and_run_clustering_experiment,
    load_and_run_forecasting_experiment,
    load_and_run_regression_experiment,
    run_classification_experiment,
    run_clustering_experiment,
    run_forecasting_experiment,
    run_regression_experiment,
)
from tsml_eval.experiments.set_classifier import get_classifier_by_name
from tsml_eval.experiments.set_clusterer import get_clusterer_by_name
from tsml_eval.experiments.set_forecaster import get_forecaster_by_name
from tsml_eval.experiments.set_regressor import get_regressor_by_name



"""Tests for building HIVE-COTE from file."""

__author__ = ["MatthewMiddlehurst"]


import numpy as np
from aeon.datasets import load_arrow_head, load_italy_power_demand
from aeon.testing.utils.estimator_checks import _assert_array_almost_equal

from tsml_eval.estimators.classification.hybrid.hivecote_from_file import (
    FromFileHIVECOTE,
)
from tsml_eval.testing.testing_utils import _TEST_RESULTS_PATH


def test_hivecote_from_file():
    """Test HIVE-COTE from file with ItalyPowerDemand results."""
    X_train, y_train = load_italy_power_demand(split="train")
    X_test, _ = load_italy_power_demand(split="test")

    file_paths = [
        _TEST_RESULTS_PATH + "/classification/Arsenal/Predictions/ItalyPowerDemand/",
        _TEST_RESULTS_PATH + "/classification/DrCIF/Predictions/ItalyPowerDemand/",
        _TEST_RESULTS_PATH + "/classification/STC/Predictions/ItalyPowerDemand/",
        _TEST_RESULTS_PATH + "/classification/TDE/Predictions/ItalyPowerDemand/",
    ]

    hc2 = FromFileHIVECOTE(classifiers=file_paths, random_state=0)
    hc2.fit(X_train, y_train)
    probas = hc2.predict_proba(X_test)

    # Component weights (a=4):
    # TDE: 0.941621858
    # STC: 0.88583781517
    # DrCIF: 0.8325698605
    # Arsenal: 0.8325698605

    # First case test probabilities:
    # TDE: 0.2091581219241968, 0.7908418780758033
    # STC: 0.025, 0.975
    # DrCIF: 0.066, 0.934
    # Arsenal: 0.0, 1.0

    # Weighted probabilities:
    # TDE: 0.19694785938, 0.74467399861
    # STC: 0.02214594537, 0.86369186979
    # DrCIF: 0.05494961079 , 0.7776202497
    # Arsenal: 0.0, 0.8325698605

    # Sum of weighted probabilities:
    # 0.27404341554, 3.2185559786

    # Normalised weighted probabilities:
    # 0.07846402768, 0.92153597231

    assert probas.shape == (X_test.shape[0], 2)
    _assert_array_almost_equal(probas[0], np.array([0.0785, 0.9215]), decimal=4)


def test_tuned_hivecote_from_file():
    """Test HIVE-COTE from file tuned alpha with ArrowHead results."""
    X_train, y_train = load_arrow_head(split="train")
    X_test, _ = load_arrow_head(split="test")

    file_paths = [
        _TEST_RESULTS_PATH + "/classification/Arsenal/Predictions/ArrowHead/",
        _TEST_RESULTS_PATH + "/classification/DrCIF/Predictions/ArrowHead/",
        _TEST_RESULTS_PATH + "/classification/STC/Predictions/ArrowHead/",
        _TEST_RESULTS_PATH + "/classification/TDE/Predictions/ArrowHead/",
    ]

    hc2 = FromFileHIVECOTE(classifiers=file_paths, tune_alpha=True, random_state=0)
    hc2.fit(X_train, y_train)
    probas = hc2.predict_proba(X_test)

    assert probas.shape == (X_test.shape[0], 3)
    _assert_array_almost_equal(probas[0], np.array([0.6092, 0.2308, 0.16]), decimal=4)



"""Abstract class for storing and loading results from an experiment."""

from abc import ABC, abstractmethod


class EstimatorResults(ABC):
    """
    Abstract class for storing and loading results from an experiment.

    Parameters
    ----------
    dataset_name : str, default="N/A"
        Name of the dataset used.
    estimator_name : str, default="N/A"
        Name of the estimator used.
    split : str, default="N/A"
        Type of data split used, i.e. "train" or "test".
    resample_id : int or None, default=None
        Random seed used for the data resample, with 0 usually being the original data.
    time_unit : str, default="nanoseconds"
        Time measurement used for other fields.
    description : str, default=""
        Additional description of the experiment. Appended to the end
        of the first line of the results file.
    parameters : str, default="No parameter info"
        Information about parameters used in the estimator and other build information.
        Written to the second line of the results file.
    fit_time : float, default=-1.0
        Time taken fitting the model.
    predict_time : float, default=-1.0
        Time taken making predictions.
    benchmark_time : float, default=-1.0
        Time taken to run a simple benchmark function. In tsml-eval experiments, this
        is the time spent to sort 1,000 (seeded) random numpy arrays of size 20,000.
    memory_usage : float, default=-1.0
        Memory usage during the experiment. In tsml-eval experiments, this is the peak
        memory usage during the fit method.
    """

    def __init__(
        self,
        dataset_name="N/A",
        estimator_name="N/A",
        split="N/A",
        resample_id=-1,
        time_unit="nanoseconds",
        description="",
        parameters="No parameter info",
        fit_time=-1.0,
        predict_time=-1.0,
        benchmark_time=-1.0,
        memory_usage=-1.0,
    ):
        # Line 1
        self.dataset_name = dataset_name
        self.estimator_name = estimator_name
        self.split = split
        self.resample_id = resample_id
        self.time_unit = time_unit
        self.description = description

        # Line 2
        self.parameter_info = parameters

        # Line 3
        self.fit_time = fit_time
        self.predict_time = predict_time
        self.benchmark_time = benchmark_time
        self.memory_usage = memory_usage

    # var_name: (display_name, higher is better, is timing)
    statistics = {
        "fit_time": ("FitTime", False, True),
        "predict_time": ("PredictTime", False, True),
        "memory_usage": ("MemoryUsage", False, False),
    }

    @abstractmethod
    def save_to_file(self, file_path, full_path=True):
        """
        Write the estimator results into a file format used by tsml.

        Abstract, must be implemented by subclasses.

        Parameters
        ----------
        file_path : str
            Path to write the results file to or the directory to build the default file
            structure if full_path is False.
        full_path : boolean, default=True
            If True, results are written directly to the directory passed in file_path.
            If False, then a standard file structure using the estimator and dataset
            names is created and used to write the results file.
        """
        pass

    @abstractmethod
    def load_from_file(self, file_path, verify_values=True):
        """
        Load estimator results from a specified file.

        This method reads a file containing estimator results and reconstructs the
        EstimatorResults object. It calculates performance statistics and
        verifies values based on the loaded data.

        Abstract, must be implemented by subclasses.

        Parameters
        ----------
        file_path : str
            The path to the file from which estimator results should be loaded. The
            file should be a tsml formatted estimator results file.
        verify_values : bool, default=True
            If the method should perform verification of the loaded values.


        Returns
        -------
        self : EstimatorResults
            The same EstimatorResults object with loaded results.
        """
        pass

    @abstractmethod
    def calculate_statistics(self, overwrite=False):
        """
        Calculate various performance statistics based on the estimator results.

        This method computes various performance metrics based on the estimators output.

        Abstract, must be implemented by subclasses.

        Parameters
        ----------
        overwrite : bool, default=False
            If the function should overwrite the current values when they are not None.
        """
        pass



"""Utility functions for results writing."""

__author__ = ["TonyBagnall", "MatthewMiddlehurst"]

__all__ = [
    "write_classification_results",
    "write_regression_results",
    "write_clustering_results",
    "write_forecasting_results",
    "write_results_to_tsml_format",
]

import os

import numpy as np


def write_classification_results(
    predictions,
    probabilities,
    class_labels,
    classifier_name,
    dataset_name,
    file_path,
    full_path=True,
    split=None,
    resample_id=None,
    time_unit="N/A",
    first_line_comment=None,
    parameter_info="No Parameter Info",
    accuracy=-1,
    fit_time=-1,
    predict_time=-1,
    benchmark_time=-1,
    memory_usage=-1,
    n_classes=-1,
    train_estimate_method="",
    train_estimate_time=-1,
    fit_and_estimate_time=-1,
):
    """Write the predictions for a classification experiment in the format used by tsml.

    Parameters
    ----------
    predictions : np.array
        The predicted values to write to file. Must be the same length as labels.
    probabilities : np.ndarray
        Estimated class probabilities. These are written after the
        predicted values for each case.
    class_labels : np.array
        The actual class values written to file with the predicted values.
    classifier_name : str
        Name of the classifier that made the predictions. Written to file and can
        determine file structure if full_path is False.
    dataset_name : str
        Name of the problem the classifier was built on.
    file_path : str
        Path to write the results file to or the directory to build the default file
        structure if full_path is False.
    full_path : boolean, default=True
        If True, results are written directly to the directory passed in file_path.
        If False, then a standard file structure using the classifier and dataset names
        is created and used to write the results file.
    split : str or None, default=None
        Either None, 'TRAIN' or 'TEST'. Influences the result file name and first line
        of the file.
    resample_id : int or None, default=None
        Indicates what random seed was used to resample the data or used as a
        random_state for the classifier.
    time_unit : str, default="N/A"
        The format used for timings in the file, i.e. 'Seconds', 'Milliseconds',
        'Nanoseconds'
    first_line_comment : str or None, default=None
        Optional comment appended to the end of the first line, i.e. the file used to
        generate the results.
    parameter_info : str, default="No Parameter Info"
        Unstructured estimator dependant information, i.e. estimator parameters or
        values from the model build.
    accuracy: float, default=-1
        The accuracy of the predictions.
    fit_time : int, default=-1
        The time taken to fit the classifier.
    predict_time : int, default=-1
        The time taken to predict the class labels.
    benchmark_time : int, default=-1
        A benchmark time for the hardware used to scale other timings.
    memory_usage : int, default=-1
        The memory usage of the classifier.
    n_classes : int, default=-1
        The number of classes in the dataset.
    train_estimate_method : str, default=""
        The method used to generate predictions for results on training data.
    train_estimate_time : int, default=-1
        The time taken to generate predictions for results on training data.
    fit_and_estimate_time : int, default=-1
        The time taken to fit the classifier to build and generate predictions for
        results on training data.

        This is not necessarily always going to be fit_time + train_estimate_time,
        i.e. if an estimate requires the model to be fit, fit_time would be
        included in the train_estimate_time value. In this case fit_time +
        train_estimate_time would time fitting the model twice.
    """
    if len(predictions) != probabilities.shape[0] != len(class_labels):
        raise IndexError(
            "The number of predicted values is not the same as the number of actual "
            "class values."
        )

    if n_classes > -1 and n_classes != probabilities.shape[1]:
        raise IndexError(
            "The number of classes is not the same as the number of probability "
            "values for each case."
        )

    third_line = (
        f"{accuracy},"
        f"{fit_time},"
        f"{predict_time},"
        f"{benchmark_time},"
        f"{memory_usage},"
        f"{n_classes},"
        f"{train_estimate_method},"
        f"{train_estimate_time},"
        f"{fit_and_estimate_time}"
    )

    write_results_to_tsml_format(
        predictions,
        class_labels,
        classifier_name,
        dataset_name,
        file_path,
        predicted_probabilities=probabilities,
        full_path=full_path,
        split=split,
        resample_id=resample_id,
        time_unit=time_unit,
        first_line_comment=first_line_comment,
        second_line=parameter_info,
        third_line=third_line,
    )


def write_regression_results(
    predictions,
    labels,
    regressor_name,
    dataset_name,
    file_path,
    full_path=True,
    split=None,
    resample_id=None,
    time_unit="N/A",
    first_line_comment=None,
    parameter_info="No Parameter Info",
    mse=-1,
    fit_time=-1,
    predict_time=-1,
    benchmark_time=-1,
    memory_usage=-1,
    train_estimate_method="",
    train_estimate_time=-1,
    fit_and_estimate_time=-1,
):
    """Write the predictions for a regression experiment in the format used by tsml.

    Parameters
    ----------
    predictions : np.array
        The predicted values to write to file. Must be the same length as labels.
    labels : np.array
        The actual label values written to file with the predicted values.
    regressor_name : str
        Name of the regressor that made the predictions. Written to file and can
        determine file structure if full_path is False.
    dataset_name : str
        Name of the problem the regressor was built on.
    file_path : str
        Path to write the results file to or the directory to build the default file
        structure if full_path is False.
    full_path : boolean, default=True
        If True, results are written directly to the directory passed in file_path.
        If False, then a standard file structure using the regressor and dataset names
        is created and used to write the results file.
    split : str or None, default=None
        Either None, 'TRAIN' or 'TEST'. Influences the result file name and first line
        of the file.
    resample_id : int or None, default=None
        Indicates what random seed was used to resample the data or used as a
        random_state for the regressor.
    time_unit : str, default="N/A"
        The format used for timings in the file, i.e. 'Seconds', 'Milliseconds',
        'Nanoseconds'
    first_line_comment : str or None, default=None
        Optional comment appended to the end of the first line, i.e. the file used to
        generate the results.
    parameter_info : str, default="No Parameter Info"
        Unstructured estimator dependant information, i.e. estimator parameters or
        values from the model build.
    mse: float, default=-1
        The mean squared error of the predictions.
    fit_time : int, default=-1
        The time taken to fit the regressor.
    predict_time : int, default=-1
        The time taken to predict the regression labels.
    benchmark_time : int, default=-1
        A benchmark time for the hardware used to scale other timings.
    memory_usage : int, default=-1
        The memory usage of the regressor.
    train_estimate_method : str, default=""
        The method used to generate predictions for results on training data.
    train_estimate_time : int, default=-1
        The time taken to generate predictions for results on training data.
    fit_and_estimate_time : int, default=-1
        The time taken to fit the regressor to build and generate predictions for
        results on training data.

        This is not necessarily always going to be fit_time + train_estimate_time,
        i.e. if an estimate requires the model to be fit, fit_time would be
        included in the train_estimate_time value. In this case fit_time +
        train_estimate_time would time fitting the model twice.
    """
    third_line = (
        f"{mse},"
        f"{fit_time},"
        f"{predict_time},"
        f"{benchmark_time},"
        f"{memory_usage},"
        f"{train_estimate_method},"
        f"{train_estimate_time},"
        f"{fit_and_estimate_time}"
    )

    write_results_to_tsml_format(
        predictions,
        labels,
        regressor_name,
        dataset_name,
        file_path,
        full_path=full_path,
        split=split,
        resample_id=resample_id,
        time_unit=time_unit,
        first_line_comment=first_line_comment,
        second_line=parameter_info,
        third_line=third_line,
    )


def write_clustering_results(
    cluster_predictions,
    cluster_probabilities,
    class_labels,
    clusterer_name,
    dataset_name,
    file_path,
    full_path=True,
    split=None,
    resample_id=None,
    time_unit="N/A",
    first_line_comment=None,
    parameter_info="No Parameter Info",
    clustering_accuracy=-1,
    fit_time=-1,
    predict_time=-1,
    benchmark_time=-1,
    memory_usage=-1,
    n_classes=-1,
    n_clusters=-1,
):
    """Write the predictions for a clustering experiment in the format used by tsml.

    Parameters
    ----------
    cluster_predictions : np.array
        The predicted values to write to file. Must be the same length as labels.
    cluster_probabilities : np.ndarray
        Estimated cluster probabilities. These are written after the predicted values
        for each case.
    class_labels : np.array
        The actual class values written to file with the predicted values. If no label
        is available for a case, a NaN value should be substituted.
    clusterer_name : str
        Name of the clusterer that made the predictions. Written to file and can
        determine file structure if full_path is False.
    dataset_name : str
        Name of the problem the clusterer was built on.
    file_path : str
        Path to write the results file to or the directory to build the default file
        structure if full_path is False.
    full_path : boolean, default=True
        If True, results are written directly to the directory passed in file_path.
        If False, then a standard file structure using the clusterer and dataset names
        is created and used to write the results file.
    split : str or None, default=None
        Either None, 'TRAIN' or 'TEST'. Influences the result file name and first line
        of the file.
    resample_id : int or None, default=None
        Indicates what random seed was used to resample the data or used as a
        random_state for the clusterer.
    time_unit : str, default="N/A"
        The format used for timings in the file, i.e. 'Seconds', 'Milliseconds',
        'Nanoseconds'
    first_line_comment : str or None, default=None
        Optional comment appended to the end of the first line, i.e. the file used to
        generate the results or a dictionary linking label indices to actual values.
    parameter_info : str, default="No Parameter Info"
        Unstructured estimator dependant information, i.e. estimator parameters or
        values from the model build.
    clustering_accuracy : float, default=-1
        The clustering accuracy of the predictions.
    fit_time : int, default=-1
        The time taken to fit the clusterer.
    predict_time : int, default=-1
        The time taken to predict the cluster labels.
    benchmark_time : int, default=-1
        A benchmark time for the hardware used to scale other timings.
    memory_usage : int, default=-1
        The memory usage of the clusterer.
    n_classes : int, default=-1
        The number of classes in the dataset.
    n_clusters : int, default=-1
        The number of clusters founds by the clusterer.
    """
    if len(cluster_predictions) != cluster_probabilities.shape[0] != len(class_labels):
        raise IndexError(
            "The number of predicted values is not the same as the number of actual "
            "class values."
        )

    if n_clusters > -1 and n_clusters != cluster_probabilities.shape[1]:
        raise IndexError(
            "The number of clusters is not the same as the number of probability "
            "values for each case."
        )

    third_line = (
        f"{clustering_accuracy},"
        f"{fit_time},"
        f"{predict_time},"
        f"{benchmark_time},"
        f"{memory_usage},"
        f"{n_classes},"
        f"{n_clusters}"
    )

    write_results_to_tsml_format(
        cluster_predictions,
        class_labels,
        clusterer_name,
        dataset_name,
        file_path,
        predicted_probabilities=cluster_probabilities,
        full_path=full_path,
        split=split,
        resample_id=resample_id,
        time_unit=time_unit,
        first_line_comment=first_line_comment,
        second_line=parameter_info,
        third_line=third_line,
    )


def write_forecasting_results(
    predictions,
    labels,
    forecaster_name,
    dataset_name,
    file_path,
    full_path=True,
    split=None,
    random_seed=None,
    time_unit="N/A",
    first_line_comment=None,
    parameter_info="No Parameter Info",
    mape=-1,
    fit_time=-1,
    predict_time=-1,
    benchmark_time=-1,
    memory_usage=-1,
):
    """Write the predictions for a forecasting experiment in the format used by tsml.

    Parameters
    ----------
    predictions : np.array
        The predicted values to write to file. Must be the same length as labels.
    labels : np.array
        The actual label values written to file with the predicted values.
    forecaster_name : str
        Name of the forecaster that made the predictions. Written to file and can
        determine file structure if full_path is False.
    dataset_name : str
        Name of the problem the forecaster was built on.
    file_path : str
        Path to write the results file to or the directory to build the default file
        structure if full_path is False.
    full_path : boolean, default=True
        If True, results are written directly to the directory passed in file_path.
        If False, then a standard file structure using the forecaster and dataset names
        is created and used to write the results file.
    split : str or None, default=None
        Either None, 'TRAIN' or 'TEST'. Influences the result file name and first line
        of the file.
    random_seed : int or None, default=None
        Indicates what random seed was used as a random_state for the forecaster.
    time_unit : str, default="N/A"
        The format used for timings in the file, i.e. 'Seconds', 'Milliseconds',
        'Nanoseconds'
    first_line_comment : str or None, default=None
        Optional comment appended to the end of the first line, i.e. the file used to
        generate the results.
    parameter_info : str, default="No Parameter Info"
        Unstructured estimator dependant information, i.e. estimator parameters or
        values from the model build.
    mape: float, default=-1
        The mean absolute percentage error of the predictions.
    fit_time : int, default=-1
        The time taken to fit the forecaster.
    predict_time : int, default=-1
        The time taken to predict the forecasting labels.
    benchmark_time : int, default=-1
        A benchmark time for the hardware used to scale other timings.
    memory_usage : int, default=-1
        The memory usage of the forecaster.
    """
    third_line = (
        f"{mape},"
        f"{fit_time},"
        f"{predict_time},"
        f"{benchmark_time},"
        f"{memory_usage}"
    )

    write_results_to_tsml_format(
        predictions,
        labels,
        forecaster_name,
        dataset_name,
        file_path,
        full_path=full_path,
        split=split,
        resample_id=random_seed,
        time_unit=time_unit,
        first_line_comment=first_line_comment,
        second_line=parameter_info,
        third_line=third_line,
    )


def write_results_to_tsml_format(
    predictions,
    labels,
    estimator_name,
    dataset_name,
    file_path,
    predicted_probabilities=None,
    full_path=True,
    split=None,
    resample_id=None,
    time_unit="N/A",
    first_line_comment=None,
    second_line="No Parameter Info",
    third_line="N/A",
):
    """Write the predictions for an experiment in the standard format used by tsml.

    Parameters
    ----------
    predictions : np.array
        The predicted values to write to file. Must be the same length as labels.
    labels : np.array
        The actual label values written to file with the predicted values. If no label
        is available for a case, a NaN value should be substituted.
    estimator_name : str
        Name of the estimator that made the predictions. Written to file and can
        determine file structure if full_path is False.
    dataset_name : str
        Name of the problem the estimator was built on.
    file_path : str
        Path to write the results file to or the directory to build the default file
        structure if full_path is False.
    predicted_probabilities : np.ndarray, default=None
        Estimated label probabilities. If passed, these are written after the
        predicted values for each case.
    full_path : boolean, default=True
        If True, results are written directly to the directory passed in file_path.
        If False, then a standard file structure using the estimator and dataset names
        is created and used to write the results file.
    split : str or None, default=None
        Either None, 'TRAIN' or 'TEST'. Influences the result file name and first line
        of the file.
    resample_id : int or None, default=None
        Indicates what random seed was used to resample the data or used as a
        random_state for the estimator.
    time_unit : str, default="N/A"
        The format used for timings in the file, i.e. 'Seconds', 'Milliseconds',
        'Nanoseconds'
    first_line_comment : str or None, default=None
        Optional comment appended to the end of the first line, i.e. the file used to
        generate the results or a dictionary linking label indices to actual values.
    second_line : str, default="No Parameter Info"
        Unstructured estimator dependant information, i.e. estimator parameters or
        values from the model build.
    third_line : str, default = "N/A"
        Summary performance information, what values are written depends on the task.
    """
    if len(predictions) != len(labels):
        raise IndexError(
            "The number of predicted values is not the same as the number of actual "
            "labels."
        )

    # If the full directory path is not passed, make the standard structure
    if not full_path:
        d = (
            ""
            if dataset_name is None or dataset_name == "" or dataset_name == "N/A"
            else f"{dataset_name}/"
        )
        file_path = f"{file_path}/{estimator_name}/Predictions/{d}"

    os.makedirs(file_path, exist_ok=True)

    if split is None:
        split = ""
    elif split.lower() != "train" and split.lower() != "test":
        raise ValueError("Unknown 'split' value - should be 'TRAIN', 'TEST' or None")

    fname = (
        f"{split.lower()}Results"
        if resample_id is None
        else f"{split.lower()}Resample{resample_id}"
    )
    fname = fname.lower() if split == "" else fname

    with open(f"{file_path}/{fname}.csv", "w") as file:
        # the first line of the output file is in the form of:
        first_line = (
            f"{dataset_name},"
            f"{estimator_name},"
            f"{'No split' if split == '' else split.upper()},"
            f"{'None' if resample_id is None else resample_id},"
            f"{time_unit.upper()},"
            f"{'' if first_line_comment is None else first_line_comment}"
        )
        file.write(first_line + "\n")

        # the second line of the output is free form and estimator-specific; usually
        # this will record info such as paramater options used, any constituent model
        # names for ensembles, etc.
        file.write(str(second_line) + "\n")

        # the third line of the file depends on the task i.e. classification or
        # regression
        file.write(str(third_line) + "\n")

        # from line 4 onwards each line should include the actual and predicted class
        # labels (comma-separated). If present, for each case, the probabilities of
        # predicting every class value for this case should also be appended to the
        # line (a space is also included between the predicted value and the
        # predict_proba). E.g.:
        #
        # if predict_proba data IS provided for case i:
        #   labels[i], preds[i],,prob_class_0[i],
        #   prob_class_1[i],...,prob_class_c[i]
        #
        # if predict_proba data IS NOT provided for case i:
        #   labels[i], preds[i]
        #
        # If labels[i] is NaN (if clustering), labels[i] is replaced with ? to indicate
        # missing
        for i in range(0, len(predictions)):
            label = "?" if np.isnan(labels[i]) else labels[i]
            file.write(f"{label},{predictions[i]}")

            if predicted_probabilities is not None:
                file.write(",")
                for j in predicted_probabilities[i]:
                    file.write(f",{j}")
            file.write("\n")



"""Tests for publication experiments estimator selection."""

import pytest

from tsml_eval.publications.y2023.tser_archive_expansion import (
    _set_tser_exp_regressor,
    expansion_regressors,
)
from tsml_eval.testing.testing_utils import _check_set_method, _check_set_method_results


def test_set_expansion_regressor():
    """Test set_tser_exp_regressor method."""
    regressor_dict = {}
    all_regressor_names = []

    _check_set_method(
        _set_tser_exp_regressor,
        expansion_regressors,
        regressor_dict,
        all_regressor_names,
    )

    _check_set_method_results(
        regressor_dict,
        estimator_name="Regressors",
        method_name="_set_tser_exp_regressor",
    )


def test_set_expansion_regressor_invalid():
    """Test set_tser_exp_regressor method with invalid estimator."""
    with pytest.raises(ValueError, match="UNKNOWN REGRESSOR"):
        _set_tser_exp_regressor("invalid")



"""Classification Experiments: code for experiments as an alternative to orchestration.

This file is configured for runs of the main method with command line arguments, or for
single debugging runs. Results are written in a standard tsml format.
"""

__author__ = ["TonyBagnall", "MatthewMiddlehurst"]

import os
import sys

os.environ["MKL_NUM_THREADS"] = "1"  # must be done before numpy import!!
os.environ["NUMEXPR_NUM_THREADS"] = "1"  # must be done before numpy import!!
os.environ["OMP_NUM_THREADS"] = "1"  # must be done before numpy import!!
os.environ["TF_NUM_INTEROP_THREADS"] = "1"
os.environ["TF_NUM_INTRAOP_THREADS"] = "1"

import numba
from aeon.utils.validation._dependencies import _check_soft_dependencies

from tsml_eval.experiments import load_and_run_classification_experiment
from tsml_eval.experiments.set_classifier import get_classifier_by_name
from tsml_eval.experiments.tests import _CLASSIFIER_RESULTS_PATH
from tsml_eval.testing.testing_utils import _TEST_DATA_PATH
from tsml_eval.utils.arguments import parse_args
from tsml_eval.utils.experiments import _results_present, assign_gpu


def run_experiment(args):
    """Mechanism for testing classifiers on the UCR data format.

    This mirrors the mechanism used in the Java based tsml. Results generated using the
    method are in the same format as tsml and can be directly compared to the results
    generated in Java.

    Attempts to avoid the use of threading as much as possible.
    """
    numba.set_num_threads(1)
    if _check_soft_dependencies("torch", severity="none"):
        import torch

        torch.set_num_threads(1)

    # if multiple GPUs are available, assign the one with the least usage to the process
    if os.environ.get("CUDA_VISIBLE_DEVICES") is None:
        try:
            gpu = assign_gpu(set_environ=True)
            print(f"Assigned GPU {gpu} to process.")  # pragma: no cover
        except Exception:
            print("Unable to assign GPU to process.")

    # cluster run (with args), this is fragile
    if args is not None and args.__len__() > 0:
        print("Input args = ", args)
        args = parse_args(args)

        # this is also checked in load_and_run, but doing a quick check here so can
        # print a message and make sure data is not loaded
        if not args.overwrite and _results_present(
            args.results_path,
            args.estimator_name,
            args.dataset_name,
            resample_id=args.resample_id,
            split="BOTH" if args.train_fold else "TEST",
        ):
            print("Ignoring, results already present")
        else:
            load_and_run_classification_experiment(
                args.data_path,
                args.results_path,
                args.dataset_name,
                get_classifier_by_name(
                    args.estimator_name,
                    random_state=(
                        args.resample_id
                        if args.random_seed is None
                        else args.random_seed
                    ),
                    n_jobs=1,
                    fit_contract=args.fit_contract,
                    checkpoint=args.checkpoint,
                    **args.kwargs,
                ),
                row_normalise=args.row_normalise,
                classifier_name=args.estimator_name,
                resample_id=args.resample_id,
                build_train_file=args.train_fold,
                write_attributes=args.write_attributes,
                att_max_shape=args.att_max_shape,
                benchmark_time=args.benchmark_time,
                overwrite=args.overwrite,
                predefined_resample=args.predefined_resample,
            )
    # local run (no args)
    else:
        # These are example parameters, change as required for local runs
        # Do not include paths to your local directories here in PRs
        # If threading is required, see the threaded version of this file
        data_path = _TEST_DATA_PATH
        results_path = _CLASSIFIER_RESULTS_PATH
        estimator_name = "ROCKET"
        dataset_name = "MinimalChinatown"
        row_normalise = False
        resample_id = 0
        train_fold = False
        write_attributes = True
        att_max_shape = 0
        benchmark_time = True
        overwrite = False
        predefined_resample = False
        fit_contract = 0
        checkpoint = None
        kwargs = {}

        classifier = get_classifier_by_name(
            estimator_name,
            random_state=resample_id,
            n_jobs=1,
            fit_contract=fit_contract,
            checkpoint=checkpoint,
            **kwargs,
        )
        print(f"Local Run of {estimator_name} ({classifier.__class__.__name__}).")

        load_and_run_classification_experiment(
            data_path,
            results_path,
            dataset_name,
            classifier,
            row_normalise=row_normalise,
            classifier_name=estimator_name,
            resample_id=resample_id,
            build_train_file=train_fold,
            write_attributes=write_attributes,
            att_max_shape=att_max_shape,
            benchmark_time=benchmark_time,
            overwrite=overwrite,
            predefined_resample=predefined_resample,
        )


if __name__ == "__main__":
    """
    Example simple usage, with arguments input via script or hard coded for testing.
    """
    print("Running classification_experiments.py main")
    run_experiment(sys.argv[1:])



"""Files for RIST pipeline publication."""

__all__ = [
    "_set_rist_classifier",
    "_set_rist_regressor",
    "_run_classification_experiment",
    "_run_regression_experiment",
    "rist_classifiers",
    "rist_regressors",
]

from tsml_eval.publications.y2023.rist_pipeline.run_classification_experiments import (
    _run_classification_experiment,
)
from tsml_eval.publications.y2023.rist_pipeline.run_regression_experiments import (
    _run_regression_experiment,
)
from tsml_eval.publications.y2023.rist_pipeline.set_rist_classifier import (
    _set_rist_classifier,
    rist_classifiers,
)
from tsml_eval.publications.y2023.rist_pipeline.set_rist_regressor import (
    _set_rist_regressor,
    rist_regressors,
)



"""Results evaluation tools."""

__all__ = [
    "evaluate_classifiers",
    "evaluate_classifiers_from_file",
    "evaluate_classifiers_by_problem",
    "evaluate_clusterers",
    "evaluate_clusterers_from_file",
    "evaluate_clusterers_by_problem",
    "evaluate_regressors",
    "evaluate_regressors_from_file",
    "evaluate_regressors_by_problem",
    "evaluate_forecasters",
    "evaluate_forecasters_from_file",
    "evaluate_forecasters_by_problem",
]

from tsml_eval.evaluation.multiple_estimator_evaluation import (
    evaluate_classifiers,
    evaluate_classifiers_by_problem,
    evaluate_classifiers_from_file,
    evaluate_clusterers,
    evaluate_clusterers_by_problem,
    evaluate_clusterers_from_file,
    evaluate_forecasters,
    evaluate_forecasters_by_problem,
    evaluate_forecasters_from_file,
    evaluate_regressors,
    evaluate_regressors_by_problem,
    evaluate_regressors_from_file,
)



"""Tests for forecasting experiments."""

import os
import runpy

import pytest
from tsml.dummy import DummyClassifier

from tsml_eval.datasets._test_data._data_sizes import DATA_TEST_SIZES
from tsml_eval.experiments import (
    forecasting_experiments,
    get_forecaster_by_name,
    run_forecasting_experiment,
    set_forecaster,
    threaded_forecasting_experiments,
)
from tsml_eval.experiments.tests import _FORECASTER_RESULTS_PATH
from tsml_eval.testing.testing_utils import (
    _TEST_DATA_PATH,
    _check_set_method,
    _check_set_method_results,
)
from tsml_eval.utils.tests.test_results_writing import _check_forecasting_file_format


def test_run_forecasting_experiment():
    """Test forecasting experiments with test data and forecasting."""
    forecaster = "NaiveForecaster"
    dataset = "ShampooSales"

    args = [
        _TEST_DATA_PATH,
        _FORECASTER_RESULTS_PATH,
        forecaster,
        dataset,
        "2",
    ]

    forecasting_experiments.run_experiment(args)

    test_file = (
        f"{_FORECASTER_RESULTS_PATH}{forecaster}/Predictions/{dataset}/"
        "testResample2.csv"
    )
    assert os.path.exists(test_file)
    _check_forecasting_file_format(
        test_file, num_results_lines=DATA_TEST_SIZES[dataset]
    )

    # test present results checking
    forecasting_experiments.run_experiment(args)

    os.remove(test_file)


def test_run_forecasting_experiment_main():
    """Test forecasting experiments main with test data and forecaster."""
    forecaster = "NaiveForecaster"
    dataset = "ShampooSales"

    # run twice to test results present check
    for _ in range(2):
        runpy.run_path(
            (
                "./tsml_eval/experiments/forecasting_experiments.py"
                if os.getcwd().split("\\")[-1] != "tests"
                else "../forecasting_experiments.py"
            ),
            run_name="__main__",
        )

    test_file = (
        f"{_FORECASTER_RESULTS_PATH}{forecaster}/Predictions/{dataset}/"
        "testResample0.csv"
    )
    assert os.path.exists(test_file)
    _check_forecasting_file_format(test_file)

    os.remove(test_file)


def test_run_threaded_forecasting_experiment():
    """Test threaded forecasting experiments with test data and forecaster."""
    forecaster = "NaiveForecaster"
    dataset = "ShampooSales"

    args = [
        _TEST_DATA_PATH,
        _FORECASTER_RESULTS_PATH,
        forecaster,
        dataset,
        "1",
        "-nj",
        "2",
        # also test normalisation and benchmark time here
        "--row_normalise",
        "--benchmark_time",
    ]

    threaded_forecasting_experiments.run_experiment(args)

    test_file = (
        f"{_FORECASTER_RESULTS_PATH}{forecaster}/Predictions/{dataset}/"
        "testResample1.csv"
    )
    assert os.path.exists(test_file)
    _check_forecasting_file_format(test_file)

    # test present results checking
    threaded_forecasting_experiments.run_experiment(args)

    # this covers the main method and experiment function result file checking
    runpy.run_path(
        (
            "./tsml_eval/experiments/threaded_forecasting_experiments.py"
            if os.getcwd().split("\\")[-1] != "tests"
            else "../threaded_forecasting_experiments.py"
        ),
        run_name="__main__",
    )

    os.remove(test_file)


def test_run_forecasting_experiment_invalid_estimator():
    """Test run_forecasting_experiment method with invalid estimator."""
    with pytest.raises(TypeError, match="forecaster must be a"):
        run_forecasting_experiment(
            [],
            [],
            DummyClassifier(),
            [],
        )


def test_get_forecaster_by_name():
    """Test get_forecaster_by_name method."""
    forecaster_lists = [
        set_forecaster.ml_forecasters,
        set_forecaster.other_forecasters,
    ]

    forecaster_dict = {}
    all_forecaster_names = []

    for forecaster_list in forecaster_lists:
        _check_set_method(
            get_forecaster_by_name,
            forecaster_list,
            forecaster_dict,
            all_forecaster_names,
        )

    _check_set_method_results(
        forecaster_dict, estimator_name="Forecasters", method_name="set_forecaster"
    )


def test_get_forecaster_by_name_invalid():
    """Test get_forecaster_by_name method with invalid estimator."""
    with pytest.raises(ValueError, match="UNKNOWN FORECASTER"):
        get_forecaster_by_name("invalid")


# def test_aeon_forecasters_available():
#     """Test all aeon forecasters are available."""
#     excluded = [
#         # composable
#         # just missing
#     ]
#
#     est = [e for e, _ in all_estimators(estimator_types="forecaster")]
#     for e in est:
#         if e in excluded:
#             continue
#
#         try:
#             assert get_forecaster_by_name(e) is not None
#         except ModuleNotFoundError:
#             continue
